# -*- coding: utf-8 -*-
import sys
import os
import subprocess
import re
from datetime import datetime, timedelta
from glob import glob
import locale
import logging
from rich.console import Console
from rich.table import Table
from rich.align import Align
from rich import print

locale.setlocale(locale.LC_ALL, "")
logger = logging.getLogger(__name__)

class MKVToolNix:

    def __init__(self, config):
        self.playlist = None
        self._srt_file = None
        self.HAS_MKVTOOLNIX = False

        self.mkvmerge = None
        self.mkvinfo = None
        self.mkvpropedit = None

        self._srt = False
        self._chapters = False
        self._cover_file = None
        self._remove_file = []

        self._cnf = config
        self._look_for_mkvtoolnix()

    @property
    def mkv_file(self):
        return self._mkv_file

    @mkv_file.setter
    def mkv_file(self, val):
        try:
            x = int(val)
            self._mkv_file = self._get_mkv_file_from_id(x)
        except ValueError:
            self._mkv_file = self._file_valid(val)
        if self._mkv_file:
            self._command = [
                self.mkvpropedit,
                self._mkv_file
            ]

    @property
    def cover_file(self):
        return self._cover_file

    @cover_file.setter
    def cover_file(self, val):
        self._cover_file = self._file_valid(val)

    @property
    def chapters(self):
        return self._chapters

    @chapters.setter
    def chapters(self, val):
        self._chapters = val

    @property
    def srt(self):
        return self._srt

    @srt.setter
    def srt(self, val):
        self._srt = val

    def execute(self, print_messages=True):
        if print_messages:
            print(f'MKV file: "{os.path.basename(self._mkv_file)}"')
            folder = os.path.dirname(self._mkv_file)
            if folder:
                print(f'  Folder: "{os.path.dirname(folder)}"')
        if self._cover_file and self._command:
            self.update_cover()

        ret = True
        if self._command and self._cover_file:
            ret = False

        if self._command and self._chapters:
            ret = False
            self.srt_to_chapters()

        # MKV shapters to SRT
        if self._srt:
            self._srt_file = self._mkv_file.replace('.mkv', '.srt')
            self.chapters_to_srt()

        if len(self._command) > 2:
            ret = self._execute()
        return ret

    def _execute(self, print_messages=True):
        # execute command
        r = subprocess.Popen(
            self._command,
            stdout=subprocess.PIPE).stdout.read()
        if print_messages:
            print(r)
        # remove temporary files
        if self._remove_file:
            for n in self._remove_file:
                os.remove(n)
        return True

    def list_mkv_files(self):
        files = self._get_mkv_file_from_id(0, return_list=True)

        if files:
            console = Console()

            table = Table(show_header=True, header_style="bold magenta")
            table.title = (
                f'List of files under [bold magenta]'
                f'{os.path.basename(self._cnf.recording_dir)}[/bold magenta]'
            )
            table.title_justify = "left"
            table.row_styles = ['', 'plum4']
            centered_table = Align.center(table)
            table.add_column("#", justify="right")
            table.add_column("Name")
            for i, n in enumerate(files):
                table.add_row(str(i+1), os.path.basename(n))
            console.print(centered_table)
        else:
            print('No recorded files found!')

    def _get_mkv_file_from_id(self, index, return_list=False, print_messages=True):
        if index >= 0:
            index -= 1
        files = glob(os.path.join(self._cnf.recording_dir, '*.mkv'))
        if files:
            files.sort()
        if return_list:
            return files
        try:
            return files[index]
        except IndexError:
            if print_messages:
                print(f'[red]Error:[/red] Index {index} not found!')
            return None

    def _file_valid(self, a_file, print_messages=True):
        if not os.path.exists(a_file):
            for n in self._cnf.recording_dir, self._cnf.data_dir:
                test_file = os.path.join(n, a_file)
                if os.path.exists(test_file):
                    return test_file
            if print_messages:
                print(f'File not found: "{a_file}"')
                sys.exit(1)
            else:
                return None
        return a_file

    def update_cover(self, print_messages=True):
        if not self.HAS_MKVTOOLNIX:
            if print_messages:
                print('  MKVToolNix not found...')
                sys.exit(1)
            else:
                return None
        if print_messages:
            print('[magenta]Setting MKV file cover image...[/magenta]')
        self._cover_file = self._file_valid(self._cover_file)
        # scan MKV file for existing cover
        r = subprocess.Popen(
            [self.mkvmerge, '-i', self._mkv_file ],
            stdout=subprocess.PIPE).stdout.read()
        s = r.decode('utf-8')
        # Validate files
        if 'container: Matroska' not in s:
            if print_messages:
                print(f'  File not supported: "{self._mkv_file}"')
                sys.exit(1)
            else:
                return None
        if not self._cover_file.endswith('.png'):
            if print_messages:
                print(f'  File not supported: "{self._cover_file}"')
                sys.exit(1)
            else:
                return None
        if print_messages:
            print(f'  (r) PNG file: "{self._cover_file}"')
        r = re.compile(r"Attachment ID ([0-9]*): type 'image/png', size [0-9]* bytes, file name 'cover'")
        cov = re.search(r, s)
        att = cov.group(1) if cov else -1
        # build command
        self._command.append('--attachment-mime-type')
        self._command.append('image/png',)
        self._command.append('--attachment-name')
        self._command.append('cover')
        if att == -1:
            self._command.append('--add-attachment')
            self._command.append(self._cover_file)
        else:
            self._command.append('--replace-attachment')
            self._command.append(str(att) + ':' + self._cover_file)

    def chapters_to_srt(self, print_messages=True):
        if print_messages:
            print('[magenta]Chapters to SRT[/magenta]')
        chapters = self._read_chapters(self._mkv_file)
        if chapters:
            out = []
            for i, n in enumerate(chapters):
                out.append(str(i+1))
                out.append(
                        n[0].replace('.', ',') + \
                        ' --> ' +
                        self._add_seconds(n[0].replace('.', ','))
                )
                out.append(n[1])
                out.append('')
            try:
                with open(self._srt_file, 'w', encoding='utf-8') as f:
                    f.write('\n'.join(out))
            except:
                if print_messages:
                    print(f'  Error writing file: "{self._srt_file}"')
                else:
                    return None
        else:
            if print_messages:
                print('  No chapters found...')
                sys.exit(1)
            else:
                return None
        if print_messages:
            print(f'  (w) SRT file: "{os.path.basename(self._srt_file)}"')
        else:
            return self._srt_file

    def _add_seconds(self, a_str, seconds=5):
        s_time, s_mili = a_str.split(',')
        t_object = datetime.strptime('2023-10-10 ' + s_time, '%Y-%m-%d %H:%M:%S')
        new_time = t_object + timedelta(seconds=seconds)
        s_time = new_time.strftime('%H:%M:%S')
        return s_time + ',' + s_mili

    def _look_for_mkvtoolnix(self):
        self.HAS_MKVTOOLNIX = False
        if sys.platform.lower().startswith('win'):
            s_path = (
                    r'C:\Program Files\MKVToolNix\mkvmerge.exe',
                    r'C:\Program Files (x86)\MKVToolNix\mkvmerge.exe',
                    os.path.join(self._cnf.stations_dir, 'mkvtoolnix', 'mkvmerge.exe')
                    )
            for n in s_path:
                if os.path.exists(n):
                    self.mkvmerge = n
                    self.HAS_MKVTOOLNIX = True
                    break
        else:
            p = subprocess.Popen(
                    'which mkvmerge',
                    shell=True,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE
                    )
            r = p.communicate()
            self.HAS_MKVTOOLNIX = p.returncode == 0
            if self.HAS_MKVTOOLNIX:
                self.mkvmerge = r[0].decode('utf-8').strip()
            if not self.HAS_MKVTOOLNIX and sys.platform.lower().startswith('dar'):
                mkvmerge_file = os.path.join(self._cnf.data_dir, 'mkvmerge')
                if os.path.exists(mkvmerge_file):
                    self.HAS_MKVTOOLNIX = True
                    self.mkvmerge = mkvmerge_file
        if self.HAS_MKVTOOLNIX:
            self.mkvinfo = self.mkvmerge.replace('mkvmerge', 'mkvinfo')
            self.mkvpropedit = self.mkvmerge.replace('mkvmerge', 'mkvpropedit')

    def _read_chapters(self, a_file, encoding='utf-8'):
        self.playlist = None
        try:
            r = subprocess.Popen(
                [self.mkvinfo, a_file ],
                stdout=subprocess.PIPE).stdout.read()
            s = r.decode(encoding)
        except:
            return None
        if s:
            if r'|+ Chapters' not in s:
                return None
            out = []
            search_str=('Chapter time start: ', 'Chapter string: ')
            for n in search_str:
                ex = re.compile(n + r'[^\n]*')
                out.append(ex.findall(s))
            out[0] = [x[:-6] for x in out[0]]
            for n in 0, 1:
                out [n] = [ x[len(search_str[n]):] for x in out[n] ]
            self.playlist = out[1][0]
            return list(zip(out[0], out[1]))
        return None

    def srt_to_chapters(self, print_messages=True):
        if print_messages:
            print('[magenta]Updating MKV chapters...[/magenta]')
        srt_file = self._mkv_file[:-4] + '.srt'
        if print_messages:
            print(f'  (r) SRT file: "{os.path.basename(srt_file)}"')
        txt_file = self._mkv_file[:-4] + '.txt'
        if os.path.exists(srt_file):
            with open(srt_file, 'r', encoding='utf-8') as f:
                lines = f.readlines()
        else:
            if print_messages:
                print(f'[red]Error:[/red] [bold magenta]SRT[/bold magents] file not found: "{srt_file}"')
                sys.exit(1)
            else:
                return None
        times = []
        titles = []
        for n in range(1,len(lines),4):
            times.append(lines[n].strip())
            titles.append(lines[n+1].strip())

        ziped = list(zip([x.split()[0].replace(',', '.') for x in times],titles))

        out = []
        outs = 'CHAPTER{0:02}={1}\nCHAPTER{0:02}NAME={2}'
        for i, n in enumerate(ziped):
            out.append(outs.format(i+1, n[0], n[1]))

        try:
            with open(txt_file, 'w', encoding='utf-8') as f:
                f.write('\n'.join(out))
        except:
            if print_messages:
                print(f'[red]Error:[/red] Cannot write [bold magenta]Chapters[/bold magents] file "{txt_file}"')
                sys.exit(1)
            else:
                return None
        self._command.append('-c')
        self._command.append(txt_file)
        self._remove_file.append(txt_file)
        return txt_file





# -*- coding: utf-8 -*-
import locale
import sys
import os
import subprocess
import shutil
import zipfile
import platform
import json
from time import sleep
import site
import glob
import re
from argparse import ArgumentParser, SUPPRESS as SUPPRESS
from shutil import which

''' This is PyRadio version this
    install.py was released for
'''
PyRadioInstallPyReleaseVersion = '0.9.3.11.31'

locale.setlocale(locale.LC_ALL, "")

try:
    import ctypes
    import win32api
    import win32ui
except ImportError:
    pass

try:
    from urllib.request import urlopen
except ImportError:
    try:
        from urllib2 import urlopen
    except ImportError:
        pass

try:
    import requests
    HAVE_REQUESTS = True
except ImportError:
    HAVE_REQUESTS = False

VERSION = ''

HAS_PIPX = bool(shutil.which('pipx'))

try:
    from rich import print
    if platform.system().lower() == 'windows':
        import psutil
except ImportError:
    if platform.system().lower() == 'windows':
        for a_module in (
                'rich',
                'psutil',
        ):
            print('Installing module: ' + a_module + ' ...')
            ret = subprocess.call('python -m pip install --upgrade ' + a_module,
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL)
            if ret != 0:
                print('\nPyRadio cannot install python module: ' + a_module)
                print('Please make sure your internet connection is up and try again')
                sys.exit(1)
        msg = '''\n\nPyRadio has installed all required python modules.
In order for them to be properly loaded, the installation script
has to start afresh.

Please execute the installation script again, like so:

python install.py

'''
        print(msg)
        sys.exit()
    else:
        print('''Error: Module "rich" not found!

Please install the above module and try again.

Debian based distros:
    sudo apt install python3-rich

Arch based distros:
    sudo pacman install python-rich

Fedora based distros:
    sudo dnf install python-rich

openSUSE based distros:
    sudo zypper install python3-rich

If everything else fails, try:
    python -m pip install rich
or
    python3 -m pip install rich
or even
    python -m pip install --user rich

''')
    sys.exit(1)

# import logging
# logger = logging.getLogger(__name__)

def print_pipx_error():
    msg = '''[red]Error:[/red] This python installation is [red]externally managed[/red], which in plain words
        means that [red]pip[/red] can only install packages within a [green]virtual environment[/green].

       The tool to do that is called [bold green]pipx[/bold green], but it is currently not installed
       in your system.

       Please install [bold green]pipx[/bold green] and try again.
'''
    print(msg)

def print_distro_packages():
    msg = '''To succesfully complete [magenta]PyRadio[/magenta]'s installation, please refer to this page:
    [plum]https://github.com/coderholic/pyradio/blob/master/build.md[/plum]
'''
    print(msg)

def is_externally_managed():
    files = [
        os.path.join(
            os.path.dirname(x),
            'EXTERNALLY-MANAGED'
        ) for x in site.getsitepackages()
    ]
    for n in files:
        if os.path.exists(n):
            return True
    return False

def print_pyradio_on():
    msg = r'''[bold magenta]
                     _____       _____           _ _
                    |  __ \     |  __ \         | (_)
                    | |__) |   _| |__) |__ _  __| |_  ___
                    |  ___/ | | |  _  // _` |/ _` | |/ _ \\
                    | |   | |_| | | \ \ (_| | (_| | | (_) |
                    |_|    \__, |_|  \_\__,_|\__,_|_|\___/
                            __/ |
                           |___/
[/bold magenta]
                               [bold]installation script[/bold]

'''
    print(msg)

def print_trying_to_install():
    print('                              trying to install for')

def find_pyradio_win_exe():
    ''' find pyradio EXE files

        Return (system_exe, user_exe)
    '''
    exe = [None, None]
    for a_path in site.getsitepackages():
        if a_path.split(os.sep)[-1].startswith('Python'):
            py_with_ver = a_path.split(os.sep)[-1]
        an_exe = os.path.join(a_path, 'Scripts' , 'pyradio.exe')
        if os.path.exists(an_exe):
            exe[0] = an_exe
    if py_with_ver:
        an_exe = os.path.join(site.getuserbase(), py_with_ver, 'Scripts' , 'pyradio.exe')
        # print('an_exe: {}'.format(an_exe))
        if os.path.exists(an_exe):
            exe[1] = an_exe
        # print('exe: {}'.format(exe))
    return exe

def fix_pyradio_win_exe():
    exe = find_pyradio_win_exe()
    if exe[0]:
        a_path = os.getenv('PROGRAMFILES(x86)')
        if a_path:
            exe[0] = exe[0].replace(a_path, '%PROGRAMFILES(x86)%')
        a_path = os.getenv('PROGRAMFILES')
        if a_path:
            exe[0] = exe[0].replace(a_path, '%PROGRAMFILES%')
    if exe[1]:
        a_path = os.getenv('APPDATA')
        if a_path:
            exe[1] = exe[1].replace(a_path, '%APPDATA%')
    return exe

def is_pyradio_user_installed():
    if platform.system().lower().startswith('darwin'):
        return False
    pyradio_path = shutil.which('pyradio')
    if not pyradio_path:
        return False
    home = os.path.expanduser('~')
    return pyradio_path.startswith(home)

def is_process_running(process_name):
    """Check if there is any running process that contains the given name, excluding the current process."""
    current_pid = os.getpid()  # Get the PID of the current process
    for proc in psutil.process_iter(['name', 'pid']):
        try:
            if (process_name.lower() in proc.info['name'].lower()) and (proc.info['pid'] != current_pid):
                return True
        except (psutil.NoSuchProcess, psutil.AccessDenied, psutil.ZombieProcess):
            pass
    return False

def isRunning():
    return
    count = 1
    ctypes.windll.kernel32.SetConsoleTitleW('PyRadio Installation')
    while is_process_running('pyradio.exe'):
        sleep(1)
        if count > 2:
            print('[bold magenta]PyRadio[/bold magenta] is still running. Please terminate it to continue ... ')
        count += 1
    print('')

def version_string_to_list(this_version):
    # logger.error('DE this_version = "{}"'.format(this_version))
    poped = False
    tokens = ('sng', 'dev', 'git')
    sp = this_version.split('-')
    while sp[-1] in tokens:
        p = sp.pop()
        if p != 'git':
            poped = True
    if poped:
        sp.pop()
    a_v = '.'.join(sp).lower()
    # a_v = this_version.replace('-', '.').lower()
    a_l = a_v.split('.')
    while len(a_l) < 4:
        a_l.append('0')
    a_n_l = []
    # logger.error('DE a_n_l = "{}"'.format(a_n_l))
    for i, n in enumerate(a_l):
        if 'beta' in n:
            a_n_l.append(-200+int(a_l[-1].replace('beta', '')))
        elif 'rc' in n:
            a_n_l.append(-100+int(a_l[-1].replace('rc', '')))
        elif 'r' in n:
            pass
        else:
            if len(n) < 4 and i > 0:
                try:
                    a_n_l.append(int(n))
                except ValueError:
                    pass
        # logger.error('DE a_n_l = "{}"'.format(a_n_l))
    return a_n_l

def get_github_long_description_for_script():
    ret = get_github_long_description()
    if ret[1]:
        print(ret[1])
    else:
        print('')

def get_github_long_description(
    only_tag_name=False,
    devel=False,
    use_sng_repo=False,
    sng_branch=False,
    do_not_exit=False
):
    ''' get PyRadio GitHub data

        Parameters
        ----------
            only_tag_name
                If True, just return the latest tag name

            use_sng_repo
                If True, user is 's-n-g'
                This means that a development release is being built

            sng_branch
                Only maked sense if use_sng_repo is True

                If False, use branch devel
                If True, use branch master

                Not Implemented yet!

        Returns
        -------
        If only_tag_name is True
            (str) tag name

        If only_tag_name is False
            (tuple) tag name, git long description

            `git long description` is equivalant to the command:
                git describe --long --tags
    '''
    user = 'coderholic'
    if only_tag_name:
        points = ('tags', )
    else:
        points = ('commits', 'tags')
        if use_sng_repo:
            user = 's-n-g'
    returns = []
    for n in points:
        ret = None
        if n == 'tags':
            url = 'https://api.github.com/repos/coderholic/pyradio/tags'
        else:
            url = 'https://api.github.com/repos/' + user + '/pyradio/' + n
            if use_sng_repo and not sng_branch:
                url += '?sha=devel'
            else:
                url += '?sha=master'
            url += '&per_page=50'
        try:
            with urlopen(url) as https_response:
                ret = https_response.read()
        except:
            if do_not_exit:
                ret = None
            else:
                print('Error: Cannot contact GitHub!\n       Please make sure your internet connection is up and try again.')
                sys.exit(1)

        try:
            returns.append(json.loads(ret))
        except:
            if do_not_exit:
                ret = None
            else:
                print('Error: Malformed GitHub response!\n       Please make sure your internet connection is up and try again.')
                sys.exit(1)

        # for r in returns:
        #     for n in r:
        #         print(n, '\n\n')

    if ret is None:
        if only_tag_name:
            return None
        return None, None

    if only_tag_name:
        return returns[0][0]['name']

    tag_hash = returns[1][0]['commit']['sha']
    tag_name = returns[1][0]['name']

    for i, n in enumerate(returns[0]):
        if n['sha'] == tag_hash:
            revision = i
            break
    else:
        revision = 0

    # print('\n\n' + tag_name)
    # print(tag_hash)
    # print(str(use_sng_repo))
    # revision=15
    # print(revision)

    if revision > 0:
        if devel:
            ''' coderholic devel branch
                currently it does not exist
            '''
            this_version = tag_name + '-r' + str(revision) + '-' + returns[0][0]['sha'][:8] + '-dev'
        else:
            ''' coderholic master branch '''
            this_version = tag_name + '-' + str(revision) + '-' + returns[0][0]['sha'][:8]
        if use_sng_repo:
            ''' sng repo '''
            this_version = tag_name + '-r' + str(revision) + '-' + returns[0][0]['sha'][:8] + '-sng'
            if not sng_branch:
                ''' sng devel branch '''
                this_version += '-dev'
    else:
        this_version = None

    # if this_version:
    #     print('this_version = ' + this_version)
    return tag_name, this_version

def get_github_tag(do_not_exit=False):
    ''' get the name of the latest PyRadio tag on GitHub '''
    return get_github_long_description(only_tag_name=True, do_not_exit=do_not_exit)

def get_next_release():
    ''' not used '''
    r = get_github_long_description()
    print(f'Description: {r}')

    sp = r[1].split('-')
    print(f'sp = {sp}')
    x = int(sp[1]) + 1
    return sp[0] + f'-{x}'

def get_devel_version():
    long_descpr = get_github_long_description(do_not_exit=True)
    if long_descpr[0]:
        if long_descpr[1]:
            return 'PyRadio ' + long_descpr[1]
        return 'PyRadio ' + long_descpr[0]
    return None

def windows_put_devel_version():
    ''' not used '''
    long_descr = get_github_long_description()[1]
    if long_descr:
        from shutil import copyfile
        cur_dir = os.getcwd()
        copyfile(os.path.join(cur_dir, 'config.py'), os.path.join(cur_dir, 'config.py.dev'))
        try:
            with open(os.path.join(cur_dir, 'config.py'), 'r', encoding='utf-8') as con:
                lines = con.read()
            lines = lines.replace("git_description = ''", "git_description = '" + long_descr + "'")
            with open(os.path.join(cur_dir, 'config.py'), 'w', encoding='utf-8') as con:
                con.write(lines)
        except:
            print('Error: Cannot change downloaded files...\n       Please close all running programs and try again.')
            sys.exit(1)

def WindowExists(title):
    ''' fixing #146  '''
    try:
        import win32ui
    except:
        pass
    try:
        win32ui.FindWindow(None, title)
    except UnboundLocalError:
        # os.system('cls')
        msg = '''\n\n[bold magenta]PyRadio[/bold magenta] has installed all required python modules.
In order for them to be properly loaded, the installation script
has to [bold green]start afresh[/bold green].

Please execute the installation script again, like so:

    [bold red]python install.py[/bold red]

'''
        print(msg)
        sys.exit()
    except win32ui.error:
        return False
    else:
        return True

def get_a_linux_resource_opener():
    ''' return an resource open tool from a list
        the list comes from "Resource openers"
        https://wiki.archlinux.org/title/default_applications

        returns
            - a list of the program's name / argument
            - None if no program is found

    '''
    programs = (
            ('xdg-open', None),
            ('gio', 'open'),
            ('mimeopen', '-n'),
            ('mimeo', None),
            ('handlr', 'open'),
            )
    for a_prog in programs:
        which_str = which(a_prog[0])
        if which_str:
            if a_prog[1]:
                return [which_str] + a_prog[1].split(' ')
            return [which_str]
    return None

def open_cache_dir():
    c = PyRadioCache()
    if not c.exists():
        print('[magenta]PyRadio Cache[/magenta]: [red]not found[/red]\n')
        sys.exit(1)
    if platform.system().lower() == 'windows':
        os.startfile(c.cache_dir)
    elif platform.system().lower() == 'darwin':
        subprocess.Popen([which('open'), c.cache_dir])
    else:
        prog = get_a_linux_resource_opener()
        if prog:
            if isinstance(prog, str):
                prog = prog.split(' ')
            try:
                subprocess.Popen(
                    [*prog, c.cache_dir],
                    stdout=subprocess.DEVNULL,
                    stderr=subprocess.DEVNULL
                )
            except (FileNotFoundError, PermissionError):
                pass
    print(f'[magenta]PyRadio[/magenta] Cache dir: "{c.cache_dir}"')


class PyRadioCache():

    _gits = (
        'pyradio-master.zip',
        'pyradio-devel.zip',
        'pyradio-master.zip'
    )

    def __init__(self):
        ''' Initialize a PyRadio Cache

            Parameters:
                if none is specified, use
                    ~/.config/pyradio/data/.cache     or
                    %APPDATA%\\pyradio\\data\\_cache
                parced in this order
            config_dir
                PyRadio config dir
                cache dir = config_dir + data + .cache
            data_dir
                PyRadio data dir
                cache_dir = data_dir + .cache
            cache_dir
                cache dir = cache_dir
        '''
        self._files = []
        if platform.system().lower().startswith('win'):
            self._cache_dir = os.path.join(
                os.path.expanduser('APPDATA'),
                'pyradio', 'data', '_cache'
            )
        else:
            cache_dir = os.getenv('XDG_CACHE_HOME')
            if cache_dir is None:
                cache_dir = os.path.join(
                    os.path.expanduser('~'),
                    '.cache', 'pyradio'
                )
            chk = (
                cache_dir,
                os.path.join(
                    os.path.expanduser('~'),
                    '.config', 'pyradio', 'data', '.cache'
                )
            )
            for n in chk:
                if os.path.exists(n):
                    self._cache_dir = n
                    break
            if self._cache_dir is None:
                self._cache_dir = chk[0]
                try:
                    os.makedirs(self._cache_dir, exist_ok=True)
                except:
                    pass
        self._del_gits()

    def list(self):
        ''' return a string
        the string is a list of file in the cache
        '''
        if self.exists():
            from rich.console import Console
            console = Console()
            if self.files:
                max_len = max(len(x) for x in self._files) + 4
                console.print('[magenta]PyRadio Cache[/magenta]:')
                for i, n in enumerate(self._files):
                    if i < 5:
                        s = os.path.join(self._cache_dir, n)
                        console.print('  [green]{0}[/green]. {1} [bold]{2}MB[/bold]'.format(i+1, n.ljust(max_len), self.get_size(s, 'mb')), highlight=False)
                    else:
                        console.print('plus [green]{}[/green] more files'.format(len(self._files)-5), highlight=False)
                        break
            else:
                print('[magenta]PyRadio Cache[/magenta]: [bold cyan]empty[/bold cyan]')
        else:
            print('[magenta]PyRadio Cache[/magenta]: [bold red]not found[/bold red]')
        print('')

    @property
    def cache_dir(self):
        ''' resturn the cache dir '''
        return self._cache_dir

    @property
    def files(self):
        ''' return the ZIP files in the cache '''
        self._read_cache()
        return self._files

    @property
    def current_version(self):
        ''' return the latest PyRadio version ZIP file name '''
        if self.files:
            return self._files[0]
        return ''

    @classmethod
    def get_size(cls, file_path, unit='bytes'):
        ''' get the size of a file and return it as
            x (bytes), x (kb), x (mb), x (gb)

            units are: 'bytes' (default), 'kb', 'mb', 'gb'
        '''
        if os.path.exists(file_path):
            file_size = os.path.getsize(file_path)
        else:
            file_size = 0
        exponents_map = {'bytes': 0, 'kb': 1, 'mb': 2, 'gb': 3}
        if unit not in exponents_map:
            raise ValueError("Must select from \
            ['bytes', 'kb', 'mb', 'gb']")

        size = file_size / 1024 ** exponents_map[unit]
        return round(size, 3)

    def delete(self):
        ''' delete the cache dir and all file in it '''
        shutil.rmtree(self._cache_dir, ignore_errors=True)

    def clear(self):
        ''' clear the cache dir
            deletes all file but the latest PyRadio ZIP version
        '''
        ret = True
        if self._files:
            del_files = [os.path.join(self._cache_dir, x) for x in self._files]
            del del_files[0]
            for n in del_files:
                try:
                    os.remove(n)
                except:
                    ret = False
        source = os.path.join(self._cache_dir, 'pyradio-source')
        if os.path.exists(source):
            shutil.rmtree(source, ignore_errors=True)
        return ret

    def exists(self):
        ''' does the cache dir exists? '''
        return os.path.exists(self._cache_dir)

    def is_empty(self):
        ''' is the cache dir empty?
            True means that either there are no ZIP files
            in the dir, or only the latest version ZIP file
            exists
        '''
        return not self.files

    def _read_cache(self):
        if self.exists():
            f = os.listdir(self._cache_dir)
            self._files = [
                x for x in f if x.endswith('.zip') and \
                    x not in self._gits
            ]
            if self._files:
                self._files.sort()
                self._files.reverse()
                #self._files.pop()
        else:
            self._files = []

        return self._files

    def _del_gits(self):
        del_files = [
            os.path.join(self._cache_dir, x) for x in self._gits
        ]
        for n in del_files:
            if os.path.exists(n):
                try:
                    os.remove(n)
                except:
                    pass


class MyArgParser(ArgumentParser):

    def print_usage(self, file=None):
        if file is None:
            file = sys.stdout
        print(self._add_colors(self.format_usage()))

    def print_help(self, file=None):
        if file is None:
            file = sys.stdout
        print(self._add_colors(self.format_help()))

    def _add_colors(self, txt):
        t = txt.replace('show this help', 'Show this help').replace('usage:', 'Usage:').replace('options:', 'Options:').replace('[', '|').replace(']', '||')
        x = re.sub(r'([^a-zZ-Z0-9])(--*[^ ,\t|]*)', r'\1[red]\2[/red]', t)
        t = re.sub(r'([A-Z_][A-Z_]+)', r'[green]\1[/green]', x)
        x = re.sub('([^"]pyradio)', r'[magenta]\1[/magenta]', t, flags=re.I)
        t = re.sub(r'(player_name:[a-z:_]+)', r'[plum2]\1[/plum2]', x)
        x = t.replace('mpv', '[green]mpv[/green]').replace('mplayer', '[green]mplayer[/green]').replace('vlc', '[green]vlc[/green]')
        return '[bold]' + x.replace('||', r']').replace('|', r'\[') + '[/bold]'


class PythonExecutable():
    ''' A class to verify that python is installed
        and in the PATH
    '''

    def __init__(
            self,
                 terminate_if_not_found=False):
        ''' Parameters
            ==========
            requested_python_version
                The version of python we are looking for
                (either 2 or 3)
            terminate_if_not_found
                If True, the program will terminate if python
                is not found (default is False)
        '''
        self.is_debian = False
        self._python = [None, None]
        self.requested_python_version = 3
        self._terminate_if_not_found = terminate_if_not_found
        if not platform.system().lower().startswith('win'):
            self._check_if_is_debian_based()
        self._get_pythons()

    def __str__(self):
        return 'Is Debian: {0}\nPython: {1}, {2}'.format(
            self.is_debian,
            self._python[0],
            self._python[1],
        )

    @property
    def python(self):
        return self._python[self.requested_python_version-2]

    @python.setter
    def python(self, value):
        raise RuntimeError('ValueError: property is read only!')

    def _check_if_is_debian_based(self):
        p = subprocess.Popen(
            'apt --version',
            shell=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE
        )
        p.communicate()
        self.is_debian = p.returncode == 0
        return self.is_debian

    def _get_pythons(self):
        ''' get python (no version) '''
        p = subprocess.Popen(
            'python --version',
            shell=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE
        )
        for com in p.communicate():
            str_com = str(com)
            if 'Python 2.' in str_com or \
                    'python 2.' in str_com:
                self._python[0] = 'python'
                break
            if 'Python 3.' in str_com or \
                    'python 3.' in str_com:
                self._python[1] = 'python'
                break

        ''' get versioned names '''
        for n in range(2, 4):
            if self._python[n-2] is None:
                self._get_python(n)

    def _get_python(self, version):
        p = subprocess.Popen(
            'python' + str(version) + ' --version',
            shell=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE
        )
        p.communicate()
        if p.returncode == 0:
            self._python[version - 2] = 'python' + str(version)

        if self._terminate_if_not_found and \
                not self.can_use() and \
                self.requested_python_version == version:
            print('''

Python {} was not found in your system.
If you have already installed it, you probably have not added it in your PATH.
To verify that Python is in your PATH open a terminal/console and type "python".
If you get en error, you have to add it to your PATH.
'''.format('2' if self.requested_python_version == 2 else '3'))
            sys.exit(1)

    def can_use(self):
        ''' Can I use the python I requested?
        '''
        return self._python[self.requested_python_version - 2]

class PyRadioUpdate():

    ''' package values:
            0   -   official release
            1   -   s-n-g release
            2   -   s-h-g devel
            3   -   official devel
    '''

    ZIP_URL = ['https://github.com/coderholic/pyradio/archive/',
               'https://github.com/s-n-g/pyradio/archive/master.zip',
               'https://github.com/s-n-g/pyradio/archive/devel.zip',
               'https://github.com/coderholic/pyradio/archive/devel.zip',
               'https://github.com/coderholic/pyradio/archive/master.zip',
               ]

    ZIP_DIR  = ['pyradio-',
                'pyradio-master',
                'pyradio-devel',
                'pyradio-devel',
                'pyradio-master'
                ]

    install = False
    user = True

    _python_exec = None

    _delete_dir_limit = 0

    _get_cache = False

    def __init__(self,
                 package=0,
                 user=True,
                 github_long_description=None,
                 pix_isolated=False
                 ):
        if platform.system().lower().startswith('win'):
            raise RuntimeError('This is a linux only class...')
        self._dir = self._install_dir = ''
        self._package = package
        self.user = user
        self._github_long_description = github_long_description
        self._python_exec = PythonExecutable(
                terminate_if_not_found=True
                )
        self._pix_isolated = pix_isolated

    def update_pyradio(self, win_open_dir=False):
        if platform.system().lower().startswith('win'):
            ''' create BAT file to update or execute '''
            if win_open_dir:
                self.update_or_uninstall_on_windows('update-open')
            else:
                self.update_or_uninstall_on_windows('update')
        else:
            ''' update PyRadio under Linux and MacOS '''
            if self._get_cache:
                print('Downloading PyRadio...')
            elif self.install:
                print('Installing PyRadio...')
            else:
                print('Updating PyRadio...')
            if self._do_it():
                if self.install:
                    print('\n\nPyRadio succesfully installed!')
                    print('Hope you have a lot of fun using it!')
                else:
                    print('\n\nPyRadio updated to the latest release!')
                    print('Thank you for your continuous support to this project!')
                print('Cheers!\n')
                sys.exit()
            else:
                print('\n\nAn error occured during the installation!')
                print('This should have never had happened...')
                print('Please report this at http://github.com/coderholic/pyradio/issues\n')
                sys.exit(1)

    def remove_pyradio(self, win_open_dir=False):
        if platform.system().lower().startswith('win'):
            ''' create BAT file to update or execute '''
            if win_open_dir:
                self.update_or_uninstall_on_windows('uninstall-open')
            else:
                self.update_or_uninstall_on_windows('uninstall')
        else:
            ''' uninstall PyRadio under Linux and MacOS '''
            print('Uninstalling PyRadio...')
            if self._do_it(mode='uninstall'):
                print('\nPyRadio has been succesfully uninstalled!')
                print('We are really sorry to see you go!')
                print('Cheers!\n')
                sys.exit()
            else:
                os.system('clear')
                print('\n\nAn error occured during the uninstallation!')
                print('This should have never had happened...')
                print('Please report this at http://github.com/coderholic/pyradio/issues\n')
                sys.exit(1)

    def update_or_uninstall_on_windows(self, mode='update', from_pyradio=False):
        # Params:
        #       mode:           the type of zip file to download
        #       from_pyradio:   True if executed by "pyradio -d"
        params = ('', '--sng-master', '--sng-devel', '--devel', '--master')
        isRunning()
        ''' Creates BAT file to update or uninstall PyRadio on Windows'''
        self._dir = os.path.join(os.path.expanduser('~'), 'tmp-pyradio')
        shutil.rmtree(self._dir, ignore_errors=True)
        os.makedirs(self._dir, exist_ok=True)
        if mode.startswith('update'):
            bat = os.path.join(self._dir, 'update.bat')
            # os.system('CLS')
            # PyRadioUpdateOnWindows.print_update_bat_created()
        else:
            bat = os.path.join(self._dir, 'uninstall.bat')
            # os.system('CLS')
            if from_pyradio:
                PyRadioUpdateOnWindows.print_uninstall_bat_created()

        if self._package == 0:
            if self.ZIP_DIR[0].endswith('-'):
                try:
                    if VERSION == '':
                        VERSION = get_github_tag()
                except:
                    VERSION = get_github_tag()
                # print('VERSION = "{}"'.format(VERSION))
                self.ZIP_URL[0] = self.ZIP_URL[0] + VERSION + '.zip'
                self.ZIP_DIR[0] += VERSION
        try:
            with open(bat, "w", encoding='utf-8') as b:
                b.write('@ECHO OFF\n')
                # b.write('CLS\n')
                b.write('python -m pip install --upgrade wheel 1>NUL 2>NUL\n')
                b.write('if %ERRORLEVEL% == 1 GOTO downloaderror\n')
                b.write('python -m pip install --upgrade setuptools 1>NUL 2>NUL\n')
                b.write('if %ERRORLEVEL% == 1 GOTO downloaderror\n')
                b.write('python -m pip install --upgrade requests 1>NUL 2>NUL\n')
                b.write('if %ERRORLEVEL% == 1 GOTO downloaderror\n')
                b.write('python -m pip install --upgrade rich 1>NUL 2>NUL\n')
                b.write('if %ERRORLEVEL% == 1 GOTO downloaderror\n')
                b.write('python -m pip install --upgrade py-cpuinfo 1>NUL 2>NUL\n')
                b.write('if %ERRORLEVEL% == 1 GOTO downloaderror\n')
                b.write('python -m pip install --upgrade winotify 1>NUL 2>NUL\n')
                b.write('if %ERRORLEVEL% == 1 GOTO downloaderror\n')
                b.write('python -m pip install --upgrade python-dateutil 1>NUL 2>NUL\n')
                b.write('if %ERRORLEVEL% == 1 GOTO downloaderror\n')
                # b.write('PAUSE\n')
                if mode.startswith('update'):
                    # it is ok, no need to compensate for zip/wheel
                    b.write('COPY "{}" . 1>NUL\n'.format(os.path.abspath(__file__)))
                    if self._package == 0:
                        b.write(self._python_exec.python + ' install.py --no-logo --do-update\n')
                    else:
                        b.write(self._python_exec.python + ' install.py --no-logo --do-update ' + params[self._package] + '\n')
                    b.write('if %ERRORLEVEL% == 1 GOTO downloaderror\n')
                    b.write('cd "' + os.path.join(self._dir, 'pyradio-source') + '"\n')

                    b.write('IF EXIST C:\\Users\\Spiros\\pyradio (\n')
                    # b.write('COPY C:\\Users\\Spiros\\pyradio\\pyradio\\install.py pyradio\n')
                    b.write('COPY C:\\Users\\Spiros\\pyradio\\pyradio\\*.py pyradio\n')
                    b.write('COPY C:\\Users\\Spiros\\pyradio\\devel\\*.bat devel\n')
                    b.write(')\n')


                    b.write('devel\\build_install_pyradio.bat -U\n')
                    b.write('GOTO endofscript\n')
                else:
                    # it is ok, no need to compensate for zip/wheel
                    b.write('COPY "{}" uninstall.py 1>NUL\n'.format(os.path.abspath(__file__)))
                    if self._package == 0:
                        b.write(self._python_exec.python + ' uninstall.py --do-uninstall\n')
                    else:
                        b.write(self._python_exec.python + ' uninstall.py --do-uninstall ' + params[self._package] + '\n')
                    b.write('if %ERRORLEVEL% == 1 GOTO downloaderror\n')
                    # print('self._dir = "{}"'.format(self._dir))
                    # print('self._package = "{}"'.format(self._package))
                    # print('self.ZIP_DIR = "{}"'.format(self.ZIP_DIR))
                    b.write('cd "' + os.path.join(self._dir, 'pyradio-source') + '"\n')

                    b.write('IF EXIST C:\\Users\\Spiros\\pyradio (\n')
                    # b.write('COPY C:\\Users\\Spiros\\pyradio\\pyradio\\install.py pyradio\n')
                    b.write('COPY C:\\Users\\Spiros\\pyradio\\pyradio\\*.py pyradio\n')
                    b.write('COPY C:\\Users\\Spiros\\pyradio\\devel\\*.bat devel\n')
                    b.write(')\n')

                    b.write('devel\\build_install_pyradio.bat -u\n')
                    # b.write('PAUSE\n')
                    b.write('GOTO endofscript\n')
                b.write('ECHO.\n\n')
                b.write(':downloaderror\n')
                # b.write('CLS\n')
                b.write('ECHO Error:\tPyRadio cannot connect to GitHub...\n')
                b.write('ECHO \tPlease make sure that your internet connection is still up and try again\n')
                b.write('ECHO.\n\n')
                b.write('PAUSE\n')
                b.write('exit 1\n')
                b.write(':endofscript\n')
                b.write('exit 0\n')
        except:
            print('\nCreating the update/uninstall BAT file failed...')
            print('You should probably reboot your machine and try again.\n')
            sys.exit(1)

        if mode.endswith('-open'):
            os.startfile(self._dir)

    def open_windows_dir(self):
        os.startfile(self._dir)

    def _no_download_method(self):
        if platform.system().lower().startswith('darwin'):
            subprocess.call('python3 -m pip install requests', shell=True)
            print('\n\nPyRadio has installed the minimum necessary modules for its execution\nPlease execute the same command again...')
        else:
            print('Error: PyRadio has no way to download files...')
            print('       Please install python\'s "requests" module and try again.\n')
        sys.exit(1)

    def _do_it(self, mode='update'):
        if not HAVE_REQUESTS:
            self._no_download_method()

        '''' get tmp dir '''
        if HAS_PIPX:
            cache_dir = os.getenv('XDG_CACHE_HOME')
            if cache_dir is None:
                self._dir = os.path.join(os.path.expanduser('~'), '.cache', 'pyradio')
            else:
                self._dir =  os.path.join(cache_dir, 'pyradio')
            if not os.path.exists(self._dir):
                self._dir = os.path.join(os.path.expanduser('~'), '.config', 'pyradio', 'data', '.cache')
        else:
            if os.path.isdir('/tmp'):
                self._dir = os.path.join('/tmp', 'tmp-pyradio')
            else:
                self._dir = os.path.join(os.path.expanduser('~'), 'tmp-pyradio')
        print(f'Using directory: "{self._dir}"')

        ''' create tmp directory '''
        self._delete_dir_limit = 0
        self._mkdir(self._dir, self._empty_dir, self._permission_error)
        if not os.path.isdir(self._dir):
            print(f'Error: Cannot create temp directory: "{self._dir}"')
            sys.exit(1)

        ''' download pyradio '''
        self._download_pyradio()

        ''' change to pyradio directory '''
        self._install_dir = os.path.join(self._dir, self._install_dir)
        if os.path.isdir(self._install_dir):
            os.chdir(self._install_dir)
            os.chmod('devel/build_install_pyradio', 0o766)
        else:
            print('Error: Failed to download PyRadio source code...\n')
            sys.exit(1)

        self._change_git_discription_in_config_py()

        param = ''
        isol = ' -i ' if self._pix_isolated else ' '
        if mode == 'update':
            ''' install pyradio '''
            if self.user:
                param += ' --user'
            comm = 'devel/build_install_pyradio' + \
                isol + ' -no-dev -x ' + \
                self._python_exec.python + ' '  + param,
            ret = subprocess.call(
                comm, shell=True
            )
        else:
            ''' uninstall pyradio '''
            ret = subprocess.call(
                'devel/build_install_pyradio -no-dev -x ' + \
                self._python_exec.python + ' -R' + param, shell=True
            )
        if ret > 0:
            ret = False
        else:
            ret = True
        self._clean_up()
        return ret

    def _change_git_discription_in_config_py(self):
        # print('\n\n_change_git_discription_in_config_py(): self._github_long_description = "{}"\n\n'.format(self._github_long_description))
        # logger.error('DE _change_git_discription_in_config_py(): self._github_long_description = "{}"'.format(self._github_long_description))
        ''' change git_discription in pyradio/config.py '''
        if self._github_long_description is not None:
            try:
                with open(os.path.join(self._install_dir, 'pyradio', 'config.py'), 'r', encoding='utf-8') as con:
                    lines = con.read()
                lines = lines.replace("git_description = ''", "git_description = '" + self._github_long_description + "'")
                with open(os.path.join(self._install_dir, 'pyradio', 'config.py'), 'w', encoding='utf-8') as con:
                    con.write(lines)
            except:
                print('Error: Cannot change downloaded files...\n       Please close all running programs and try again.')
                sys.exit(1)

    def _download_pyradio(self):
        os.chdir(self._dir)
        if self._package == 0:
            try:
                VERSION ==  ''
            except:
                VERSION = get_github_tag()
            self.ZIP_URL[0] = self.ZIP_URL[0] + VERSION + '.zip'
            self.ZIP_DIR[0] += VERSION
        print('Downloading PyRadio source code...')
        self._install_dir = self.ZIP_DIR[self._package]
        if not self._download_file(self.ZIP_URL[self._package],
                os.path.join(self._dir, self.ZIP_DIR[self._package] + '.zip')):
            print('Error: Failed to download PyRadio source code...\n')
            sys.exit(1)

        print('Extracting PyRadio source code...')
        with zipfile.ZipFile(os.path.join(self._dir, self.ZIP_DIR[self._package] + '.zip')) as z:
            try:
                z.extractall(path=self._dir)
            except:
                print('Error: PyRadio source code ZIP file is corrupt...\n')
                sys.exit(1)
        try:
            os.rename(
                os.path.join(self._dir, self.ZIP_DIR[self._package]),
                os.path.join(self._dir, 'pyradio-source')
            )
            self.ZIP_DIR[self._package] = 'pyradio-source'
            self._install_dir = 'pyradio-source'
        except:
            print('Error creating source code directory!')
            self._clean_up()
            sys.exit(1)

        if self._get_cache:
            sys.exit()

        with open(os.path.join(self._dir, self.ZIP_DIR[self._package], 'DEV'), 'w', encoding='utf-8'):
            pass
        # input('Please update files as needed. Then press ENTER to continue...')
        # sys.exit()

    def _mkdir(self, name, dir_exist_function=None, _permission_error_function=None):
        if os.path.isdir(name):
            self._clean_up()
        try:
            os.makedirs(name)
        except PermissionError:
            if _permission_error_function:
                _permission_error_function(name)
            else:
                print('Insufficient permissions...')
        except FileExistsError:
            if dir_exist_function:
                dir_exist_function(name)
        else:
            print('Dir already exists...')

    def _empty_dir(self, name=None):
        ddir = self._dir if name is None else name
        print(f'Old "{ddir}" found. Deleting...')
        self._delete_dir_limit += 1
        if self._delete_dir_limit > 2:
            print('\n\nError: Cannot delete directory: "' + ddir  +  '"')
            print('       Please relete it manually and try again... ')
            sys.exit(1)
        if HAS_PIPX:
            r_dir = os.path.join(ddir, 'pyradio-source')
            shutil.rmtree(r_dir, ignore_errors=True)
            if os.path.exists(r_dir):
                print(f'[red]Error:[/red] Cannot remove "{r_dir}"')
                print('       Please close all open programs and try again...')
                sys.exit(1)
        else:
            shutil.rmtree(ddir, ignore_errors=True)
            self._mkdir(ddir, self._empty_dir, self._permission_error)

    def _permission_error(self):
        print(f'Error: You don\'t have permission to create: "{self._dir}\"\n')
        sys.exit(1)

    def _clean_up(self):
        if not HAS_PIPX:
            shutil.rmtree(self._dir, ignore_errors=True)

    def _prompt_sudo(self):
        ret = 0
        if os.geteuid() != 0:
            msg = "[sudo] password for %u: "
            try:
                ret = subprocess.check_call(f"sudo -v -p '{msg}'", shell=True)
                return ret
            except subprocess.CalledProcessError:
                print('\nError: You must be root to execute this script...\n')
                sys.exit(1)

    def _download_file(self, url, filename):
        print(f'  url: "{url}"')
        print(f'  filename: "{filename}"')
        if os.path.exists(filename) and not (
                filename.endswith('-master.zip') or \
                filename.endswith('-devel.zip')
                ):
            print('  [magenta]** file found in cache![/magenta]')
        else:
            try:
                r = requests.get(url)
            except:
                return False
            try:
                with open(filename, 'wb') as f:
                    for chunk in r.iter_content(chunk_size=1024):
                        if chunk: # filter out keep-alive new chunks
                            f.write(chunk)
                            f.flush()
                            os.fsync(f.fileno())
            except:
                return False
        return True

class PyRadioUpdateOnWindows(PyRadioUpdate):

    def __init__(self,
                 fromTUI=False,
                 package=0,
                 github_long_description=None,
                 pix_isolated=False
                 ):
        if not platform.system().lower().startswith('win'):
            raise RuntimeError('This is a windows only class...')
        self._dir = os.path.join(os.path.expanduser('~'), 'tmp-pyradio')
        self._package = package
        self._fromTUI = fromTUI
        self._github_long_description = github_long_description
        self._python_exec = PythonExecutable(
                terminate_if_not_found=True
                )
        self._pix_isolated = pix_isolated
        self._get_cache = False

    @classmethod
    def print_update_bat_created(cls):
        print('To complete the process you will have to execute a batch file.')
        print('Windows Explorer will open the location of the batch file to run.\n')
        print('Please double click\n\n    update.bat\n\nto get PyRadio\'s latest version.\n')
        print('After you are done, you can delete the folder:\n\n    "{}"'.format(os.path.join(os.path.expanduser("~"), 'tmp-pyradio')))

    @classmethod
    def print_uninstall_bat_created(cls):
        print('To complete the process you will have to execute a batch file.')
        print('Windows Explorer will open the location of the batch file to run.\n')
        print('Please double click\n\n    uninstall.bat\n\nto remove PyRadio from your system.\n')
        print('After you are done, you can delete the folder:\n\n    "{}"'.format(os.path.join(os.path.expanduser("~"), 'tmp-pyradio')))

    def update_pyradio(self):
        self._do_it()

    def remove_pyradio(self):
        self._get_cache = False
        self._do_it(mode='uninstall')

    def _do_it(self, mode='update'):
        if not HAVE_REQUESTS:
            self._no_download_method()

        self._download_pyradio()

        ''' change to pyradio directory '''
        self._install_dir = os.path.join(self._dir, self._install_dir)
        if os.path.isdir(self._install_dir):
            os.chdir(self._install_dir)
        else:
            print('Error: Failed to download PyRadio source code...\n')
            sys.exit(1)

        os.chdir(self._dir)

        self._change_git_discription_in_config_py()


if __name__ == '__main__':
    parser = MyArgParser(
        description='Curses based Internet Radio Player',
        epilog='When executed without an argument, it installs PyRadio (stable release).'
    )
    # parser = ArgumentParser(description='PyRadio update / uninstall tool',
    #                         epilog='When executed without an argument, it installs PyRadio (stable release).')
    parser.add_argument('-U', '--update', action='store_true',
                        help='Update PyRadio.')
    parser.add_argument('-f', '--force', action='store_true',
                        help='Force installation (even if already installed).')
    if not platform.system().lower().startswith('win'):
        if HAS_PIPX:
            parser.add_argument('-i', '--isolate', action='store_true',
                                help='Install using pipx in an fully Isolated Environment (all dependencies will be installed in a virtual environment); the default is to have pipx depend on distro python packages.')
            parser.add_argument('-oc', '--open-cache', action='store_true',
                               help='Open the Cache folder')
            parser.add_argument('-sc', '--show-cache', action='store_true',
                               help='Show Cache contents')
            parser.add_argument('-cc', '--clear-cache', action='store_true',
                               help='Clear Cache contents')
        else:
            parser.add_argument('-oc', '--open-cache', action='store_true', help=SUPPRESS)
            parser.add_argument('-sc', '--show-cache', action='store_true', help=SUPPRESS)
            parser.add_argument('-cc', '--clear-cache', action='store_true', help=SUPPRESS)
            parser.add_argument('-i', '--isolate', action='store_true', help=SUPPRESS)
        if HAS_PIPX:
            parser.add_argument('-gc', '--get-cache', action='store_true',
                                help='Download source code, keep it in the cache and exit.')
        else:
            parser.add_argument('-gc', '--get-cache', action='store_true', help=SUPPRESS)
    else:
        parser.add_argument('-oc', '--open-cache', action='store_true', help=SUPPRESS)
        parser.add_argument('-sc', '--show-cache', action='store_true', help=SUPPRESS)
        parser.add_argument('-cc', '--clear-cache', action='store_true', help=SUPPRESS)
        parser.add_argument('-i', '--isolate', action='store_true', help=SUPPRESS)
    parser.add_argument('-R', '--uninstall', action='store_true',
                        help='Uninstall PyRadio.')

    ''' to be used by intermediate scripts '''
    parser.add_argument('--do-update', action='store_true', help=SUPPRESS)
    parser.add_argument('--do-uninstall', action='store_true', help=SUPPRESS)

    ''' extra downloads
        only use them after the developer says so,
        for debug purposes only
            --devel         download official devel branch
            --sng-master    download developer release (master)
            --sng-devel     download developer devel branch
            --force         force installation (even if already installed)
    '''
    parser.add_argument(
        '--git', action='store_true',
        help='install master branch from github (latest unreleased).')
    parser.add_argument('--master', action='store_true', help=SUPPRESS)
    parser.add_argument('--devel', action='store_true', help=SUPPRESS)
    parser.add_argument('--sng-master', action='store_true', help=SUPPRESS)
    parser.add_argument('--sng-devel', action='store_true', help=SUPPRESS)
    parser.add_argument('--no-logo', action='store_true', help=SUPPRESS)
    parser.add_argument('--first', action='store_true', help=SUPPRESS)
    parser.add_argument('-v', '--version', action='store_true',
                        help='Print the version of the script')

    args = parser.parse_args()
    sys.stdout.flush()

    if args.open_cache:
        open_cache_dir()
        sys.exit()

    if args.show_cache:
        c = PyRadioCache()
        c.list()
        sys.exit()

    if args.clear_cache:
        c = PyRadioCache()
        if c.exists():
            if len(c.files) > 1:
                c.clear()
            print('[magenta]PyRadio Cache[/magenta]: [green]cleared[/green]\n')
            sys.exit()
        c.list()
        sys.exit(1)

    if args.version:
        print(f'[bold green]install.py[/bold green]: installation script for [bold magenta]PyRadio {PyRadioInstallPyReleaseVersion}[/bold magenta]\n')
        sys.exit()

    if not platform.system().lower().startswith('win'):
        if os.getuid() == 0 or os.getgid() == 0:
            print('[red]Error:[/red] You must not run this script as [green]root[/green]\n')
            sys.exit(1)

    if is_externally_managed() and \
            not HAS_PIPX:
        print_pipx_error()
        print_distro_packages()
        sys.exit()
    use_logo = not args.no_logo

    if use_logo and not args.open_cache:
        print_pyradio_on()

    python_exec = PythonExecutable(
            terminate_if_not_found=True
            )

    if not python_exec.can_use:
        print('Error: Python was not found on your system...\n')
        sys.exit(1)

    ''' download official release '''
    package = 0
    tag_name = github_long_description = None
    if args.sng_master:
        ''' sng master '''
        args.force = True
        package = 1
        VERSION, github_long_description = get_github_long_description(use_sng_repo=True, sng_branch=True)
        # if not github_long_description:
        #     github_long_description = 'PyRadio'
        # if github_long_description:
        #     github_long_description = github_long_description.replace('-', '-r', 1)
        # github_long_description += '-sng'
    elif args.sng_devel:
        '''' sng devel '''
        args.force = True
        package = 2
        VERSION, github_long_description = get_github_long_description(use_sng_repo=True)
        # if not github_long_description:
        #     github_long_description = 'PyRadio'
        # id github_long_description:
        #     github_long_description = github_long_description.replace('-', '-r', 1)
        # github_long_description += '-sng-dev'
    elif args.devel:
        ''' official devel '''
        package = 3
        ''' go back to master '''
        args.force = True
        package = 0
        VERSION = get_github_tag()
    elif args.master or args.git:
        ''' official master '''
        args.force = True
        package = 4
        VERSION, github_long_description = get_github_long_description()
    else:
        VERSION = get_github_tag()

    if VERSION is None:
        VERSION = PyRadioInstallPyReleaseVersion

    if args.uninstall:
        if platform.system().lower().startswith('win'):
            ''' ok, create BAT file on Windows'''
            uni = PyRadioUpdateOnWindows(package=package)
            uni.update_or_uninstall_on_windows(mode='uninstall-open')
            uni.print_uninstall_bat_created()
        else:
            uni = PyRadioUpdate(package=package)
            uni.remove_pyradio()
        sys.exit()
    elif args.update:
        if platform.system().lower().startswith('win'):
            ''' ok, create BAT file on Windows'''
            upd = PyRadioUpdateOnWindows(
                package=package,
                github_long_description=github_long_description,
                pix_isolated=args.isolate
            )
            upd.update_or_uninstall_on_windows(mode='update-open')
            upd.print_update_bat_created()
        else:
            upd = PyRadioUpdate(
                package=package,
                github_long_description=github_long_description,
                pix_isolated=args.isolate
            )
            if args.get_cache and HAS_PIPX:
                upd._get_cache = True
            upd.user = is_pyradio_user_installed()
            upd.update_pyradio()
        sys.exit()
    elif args.do_uninstall:
        ''' coming from uninstall BAT file on Windows'''
        uni = PyRadioUpdateOnWindows(
            package=package,
            pix_isolated=args.isolate
        )
        uni.remove_pyradio()
        sys.exit()
    elif args.do_update:
        ''' coming from update BAT file on Windows'''
        upd = PyRadioUpdateOnWindows(
            package=package,
            github_long_description=github_long_description,
            pix_isolated=args.isolate
        )
        upd.update_pyradio()
        sys.exit()

    ''' Installation!!! '''
    if platform.system().lower().startswith('win'):
        exe = find_pyradio_win_exe()
        first_install = False
        if exe == [None, None]:
            first_install = True

        if not args.force:
            # is pyradio.exe in PATH
            ret = subprocess.call('pyradio -h 1>NUL 2>NUL', shell=True)
            if ret == 0 or not first_install:
                print('PyRadio is already installed.\n')
                sys.exit(1)
        for a_module in (
                'setuptools',
                'windows-curses',
                'pywin32',
                'dnspython',
                'requests',
                'rich',
                'python-dateutil',
                'psutil',
                'patool',
                'pyunpack',
                'wheel',
                'winotify',
                'py-cpuinfo',
        ):
            print('Checking module: [bold green]' + a_module + '[/bold green] ...')
            ret = subprocess.call('python -m pip install --upgrade ' + a_module,
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL)
            if ret != 0:
                print('\nPyRadio cannot install python module: ' + a_module)
                if a_module == 'windows-curses':
                    print('''This means that either you internet connection has failed
(in which case you should fix it and try again), or that
curses packagers have not yet produced a package for this
version of python (it was probably released recently).

What can you do?
1. Wait for the package to be updated (which means you will
    not be able to use PyRadio until then), or
2. Uninstall python and then go to
          https://www.python.org/downloads/
    and download and install the second to last version.
.
Then try installing PyRadio again
''')
                else:
                    print('Please make sure your internet connection is up and try again')
                sys.exit(1)

        uni = PyRadioUpdateOnWindows(
            package=package,
            github_long_description=github_long_description,
            pix_isolated=args.isolate
        )
        uni.update_or_uninstall_on_windows(
            mode='update-open'
        )
        while not os.path.isfile(os.path.join(uni._dir, 'update.bat')):
            pass

        os.chdir(uni._dir)
        if subprocess.call('update.bat') == 0:
            if first_install:
                files = glob.glob(uni._dir + '\\**\\win.py', recursive=True)
                for n in files:
                    if '\\build\\' not in n:
                        win_file = n
                        break
                subprocess.call('python "' + win_file + '"')
            print('\n\nNow you can delete the folder:')
            print(f'    "{uni._dir}"')
            print('and the file:')
            # it is ok, no need to compensate for zip/wheel
            print(f'    "{__file__}"')
    else:
        if not args.force and not args.get_cache:
            ret = subprocess.call('pyradio -h 1>/dev/null 2>&1', shell=True)
            if ret == 0:
                print('PyRadio is already installed.\n')
                sys.exit(1)
        uni = PyRadioUpdate(
            package=package,
            github_long_description=github_long_description,
            pix_isolated=args.isolate
        )
        uni.install = True
        # if not platform.system().lower().startswith('darwin'):
        #     uni.user = True
        if args.get_cache:
            uni._get_cache = True
        uni.update_pyradio()



# -*- coding: utf-8 -*-
import os
from sys import platform

import locale
locale.setlocale(locale.LC_ALL, "")

def RemoveWinVlcLogFiles(*args):
    ''' Removes all VLC log files within pyradio config
        directory on Windows.

        Files currently in use will not be deleted.
    '''
    if platform.startswith('win'):
        adir = args[0]
        # print('config = "{}"'.format(adir))
        files = [file for file in os.listdir(adir) if 'vlc_log.' in file]
        if files:
            for afile in files:
                #i print(afile)
                try:
                    # print('removing "{}"'.format(afile))
                    os.remove(os.path.join(adir, afile))
                except:
                    pass


if __name__ == "__main__":
    # example:
    import threading
    threading.Thread(target=RemoveWinVlcLogFiles('C:\\Users\\Spiros\\AppData\\Roaming\\pyradio')).start()



# -*- coding: utf-8 -*-
import subprocess
from sys import platform

import locale
locale.setlocale(locale.LC_ALL, "")

def ping(server, count=10, timeout_in_seconds=1):
    ''' ping a server on any platform
        Returns:
             1: server is alive (True)
             0: server is not alive (False)
            -1: error
    '''
    if platform.lower().startswith('win'):
        return windows_ping(server, count=count, timeout_in_miliseconds=timeout_in_seconds * 1000)
    return linux_ping(server, count=count, timeout_in_seconds=timeout_in_seconds)

def windows_ping(server, count=1, timeout_in_miliseconds=1000):
    ''' ping a server on windows
    Returns:
             1: server is alive (True)
             0: server is not alive (False)
            -1: error
    '''
    try:
        r=subprocess.Popen(
            ['ping', '-n', str(count), '-w',
             str(timeout_in_miliseconds), server ],
            stdout=subprocess.PIPE).stdout.read()
        return 0 if '100%' in str(r) else 1
    except:
        return -1

def linux_ping(server, count=1, timeout_in_seconds=1):
    ''' ping a server on linux
    Returns:
             1: server is alive (True)
             0: server is not alive (False)
            -1: error
    '''
    try:
        r = subprocess.Popen(
            ['ping', '-c', str(count), '-w',
             str(timeout_in_seconds), server],
            stderr=subprocess.DEVNULL,
            stdout=subprocess.PIPE
        ).stdout.read()
        return 0 if '100%' in str(r) else 1
    except:
        return -1

if __name__ == "__main__":
    msg= ''' ping a server on any platform
    Returns:
         1: server is alive (True)
         0: server is not alive (False)
        -1: error
    '''
    print(msg)
    ping_result = ping('www.google.com', count=20, timeout_in_seconds=3)
    print(f'Ping response: {ping_result}')



# -*- coding: utf-8 -*-
import os
import sys
import json
import queue
import socket
import threading
import subprocess
import logging

from .os_media_base import MediaControls


logger = logging.getLogger("pyradio.macos_media")

HELPER_MODULE = "pyradio.macos_nowplaying_helper"
SOCKET_PATH = "/tmp/pyradio-nowplaying-helper.sock"


class MacOSMediaController(MediaControls):

    def __init__(self, identity="PyRadio", instance_name=None, default_icon=None):
        MediaControls.__init__(self, identity, instance_name, default_icon)

        self._proc = None
        self._cmdq = queue.Queue()
        self._reader_thread = None
        self._running = False
        self._socket_path = SOCKET_PATH

    # ----------------------------------------------------------
    # lifecycle
    # ----------------------------------------------------------

    def start(self):
        try:
            if self._running:
                return

            cmd = [
                sys.executable,
                "-m",
                HELPER_MODULE,
                "run",
                "--icon",
                self._get_default_icon()
            ]

            self._proc = subprocess.Popen(
                cmd,
                stdin=subprocess.DEVNULL,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                text=True,
                bufsize=1
            )

            self._reader_thread = threading.Thread(
                target=self._reader_loop,
                name="macos-helper-reader",
                daemon=True
            )
            self._reader_thread.start()

            if not self._wait_for_helper_socket():
                self._running = False
                return

            self._running = True

        except Exception as e:
            logger.error("OS-MEDIA: MAC: Failed starting macOS helper: %s", e)

    def stop(self):
        try:
            if not self._running:
                return

            self._send_command({"type": "shutdown"})

            try:
                self._proc.wait(timeout=3)
            except Exception:
                try:
                    self._proc.terminate()
                except Exception as e:
                    logger.error("OS-MEDIA: MAC: Failed terminating helper: %s", e)

            self._running = False

        except Exception as e:
            logger.error("OS-MEDIA: MAC: Failed stopping macOS helper: %s", e)

    # ----------------------------------------------------------
    # poll bridge
    # ----------------------------------------------------------

    def poll(self, enabled):
        try:
            while True:
                msg = self._cmdq.get_nowait()

                if msg.get("type") != "cmd":
                    continue

                name = msg.get("name")

                if name == "play" and self.cb_play:
                    self.cb_play()

                elif name == "stop" and self.cb_stop:
                    self.cb_stop()

                elif name == "next" and self.cb_next:
                    self.cb_next()

                elif name == "prev" and self.cb_prev:
                    self.cb_prev()

                elif name == "playpause" and self.cb_playpause:
                    self.cb_playpause()

                elif name == "set_volume" and self.cb_set_volume:
                    self.cb_set_volume(msg.get("value"))

        except queue.Empty:
            pass
        except Exception as e:
            logger.error("OS-MEDIA: MAC: poll dispatch failed: %s", e)

    # ----------------------------------------------------------
    # updates from PyRadio
    # ----------------------------------------------------------

    def _wait_for_helper_socket(self, timeout=3.0):
        import time

        end_time = time.time() + timeout

        while time.time() < end_time:
            if self._proc is not None and self._proc.poll() is not None:
                logger.error("OS-MEDIA: MAC: macOS helper exited during startup")
                return False

            if os.path.exists(self._socket_path):
                try:
                    sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
                    sock.connect(self._socket_path)
                    sock.close()
                    return True
                except Exception:
                    pass

            time.sleep(0.05)

        logger.error("OS-MEDIA: MAC: timed out waiting for helper socket readiness")
        return False

    def update_playback(self, playing):
        self._send_command({
            "type": "playback",
            "playing": bool(playing)
        })

    def update_nav_caps(self, can_next, can_prev):
        self._send_command({
            "type": "nav_caps",
            "can_next": bool(can_next),
            "can_prev": bool(can_prev)
        })

    def update_metadata(
        self,
        trackid,
        title,
        artist,
        album,
        url=None,
        art_url=None
    ):
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug(f'OS_MEDIA: MAC: {trackid = }')
            logger.debug(f'OS-MEDIA: MAC: MAC: {title = }')
            logger.debug(f'OS-MEDIA: MAC: MAC: {artist = }')
            logger.debug(f'OS-MEDIA: MAC: MAC: {album = }')
            logger.debug(f'OS-MEDIA: MAC: MAC: {url = }')
            logger.debug(f'OS-MEDIA: MAC: MAC: {art_url = }')
        self._send_command({
            "type": "metadata",
            "trackid": trackid,
            "title": title,
            "artist": artist,
            "album": album,
            "url": url,
            "art_url": art_url
        })

    # ----------------------------------------------------------

    def make_trackid(self, *parts):
        try:
            out = []

            for part in parts:
                if part is None:
                    continue

                text = str(part).strip()
                if text:
                    out.append(text)

            if not out:
                return "pyradio:track:unknown"

            return "pyradio:track:" + "|".join(out)

        except Exception as e:
            logger.error("OS-MEDIA: MAC: MAC: Failed creating track id: %s", e)
            return "pyradio:track:unknown"

    # ----------------------------------------------------------
    # IPC
    # ----------------------------------------------------------

    def _send_command(self, payload):
        import errno
        import time

        if not self._running:
            return

        last_error = None

        for _ in range(10):
            try:
                if self._proc is not None and self._proc.poll() is not None:
                    logger.error("OS-MEDIA: MAC: helper process is not running")
                    return

                sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
                sock.connect(self._socket_path)

                data = json.dumps(payload).encode("utf-8") + b"\n"
                sock.sendall(data)
                sock.close()
                return

            except OSError as e:
                last_error = e

                if e.errno in (2, 61):
                    time.sleep(0.05)
                    continue

                logger.error("OS-MEDIA: MAC: IPC send failed: %s", e)
                return

            except Exception as e:
                logger.error("OS-MEDIA: MAC: IPC send failed: %s", e)
                return

        if last_error is not None:
            logger.error("OS-MEDIA: MAC: IPC send failed after retries: %s", last_error)

    # ----------------------------------------------------------
    # reader
    # ----------------------------------------------------------

    def _reader_loop(self):
        try:
            while True:
                line = self._proc.stdout.readline()

                if not line:
                    break

                line = line.strip()

                if not line:
                    continue

                try:
                    msg = json.loads(line)
                except Exception:
                    continue

                self._cmdq.put(msg)

        except Exception as e:
            logger.error("OS-MEDIA: MAC: Helper reader loop failed: %s", e)

    # ----------------------------------------------------------

    def _get_default_icon(self):
        try:
            base = os.path.dirname(os.path.abspath(__file__))
            return os.path.join(base, "data", "pyradio.png")
        except Exception as e:
            logger.error("OS-MEDIA: MAC: Failed resolving default icon: %s", e)
            return ""



# -*- coding: utf-8 -*-
import logging
import locale
import io
import csv
import curses
import hashlib
from os import rename, remove, access, X_OK, getenv, makedirs
from os.path import exists, dirname, join, expanduser, basename
from shutil import which, move, Error as shutil_Error
from enum import IntEnum
from sys import platform
from pathlib import Path
from rich import print
try:
    # Python ≥ 3.9
    from importlib.resources.abc import Traversable
except ImportError:
    # Python 3.7 & 3.8 (backport)
    from importlib_resources.abc import Traversable

logger = logging.getLogger(__name__)

locale.setlocale(locale.LC_ALL, "")

""" Theming constants """
def FOREGROUND():
    return 0
def BACKGROUND():
    return 1

# for pop up window
CAPTION = 2
BORDER = 3

class Station(IntEnum):
    name = 0
    url = 1
    encoding =2
    icon = 3
    profile = 4
    buffering = 5
    http = 6
    volume = 7
    referer = 8
    player = 9

M_STRINGS = {
	'checking-playlist': ' (Checking Playlist)',
	'session-locked': ' (Session Locked)',
	'session-locked-title': 'Session Locked',
	'win-title': 'Your Internet Radio Player',
	'init_': 'Initialization: ',
	'connecting_': 'Connecting to: ',
	'playing_': 'Playing: ',
	'buffering_': 'Buffering: ',
	'station_': 'Station: ',
	'station_buffering': 'Station Buffering',
	'station-open': ' - Opening connection...',
	'selected_player_': 'Selected player: ',
	'down-icon': 'Downloading icon...',
	'player-acivated_': ': Player activated!!!',
	'hist-empty': 'History is empty!!!',
	'hist-first': 'Already at first item!!!',
	'hist-last': 'Already at last item!!!',
    'muted': '[Muted] ',
    'title_': 'Title: ',
    'player-stopped': 'Player is stopped!',
    'plb-stopped': 'Playback stopped',
    'html-player-stopped': '<div class="alert alert-danger">Player is <b>stopped!</b></div>',
	'press-?': ' Press ? for help',
	'error-str': 'error',
	'vol_': 'Vol: ',
    'volume_set': 'Volume set to ',
    'error-403': 'Server returned "Forbidden" (error 403)',
    'error-404': 'Station does not exist (error 404)',
    'error-503': 'Service not available (error 503)',
    'error-1000': 'Player terminated abnormally! (error 1000)',
    'error-1001': 'Connection failed (error 1001)',
    'error-1002': 'No stream found (error 1002)',
    'error-1003': 'Connection refused (error 1003)',
    'error-1004': 'Unrecognized file format (error 1004)',
    'error-1005': 'DNS Resolution failure (error 1005)',
    'error-1006': 'Server is unreachable (error 1006)',
    'error-1007': 'Permission denied (error 1007)',
    'error-1008': 'Unrecognized file format (error 1008)',
}

""" Messages to display when player starts / stops
    Used in log to stop runaway threads from printing
    messages after playback is stopped """
player_start_stop_token = {
    0:       M_STRINGS['init_'],
    1:       M_STRINGS['plb-stopped'],
    3:       M_STRINGS['player-acivated_'],
    403:     M_STRINGS['error-403'],
    404:     M_STRINGS['error-404'],
    503:     M_STRINGS['error-503'],
    1000:    M_STRINGS['error-1000'],
    1001:    M_STRINGS['error-1001'],
    1002:    M_STRINGS['error-1002'],
    1003:    M_STRINGS['error-1003'],
    1004:    M_STRINGS['error-1004'],
    1005:    M_STRINGS['error-1005'],
    1006:    M_STRINGS['error-1006'],
    1007:    M_STRINGS['error-1007'],
    1008:    M_STRINGS['error-1008'],
}

seconds_to_KB_128 = (
    0, 78, 93, 109, 125, 140, 156, 171, 187, 203, 218, 234,
    250, 265, 281, 296, 312, 328, 343, 359, 375, 390, 406,
    421, 437, 453, 468, 484, 500, 515, 531, 546, 562, 578,
    593, 609, 625, 640, 656, 671, 687, 703, 718, 734, 750,
    765, 781, 796, 812, 828, 843, 859, 875, 890, 906, 921, 937
)

seconds_to_KB_192 = (
    0, 117, 140, 164, 187, 210, 234, 257, 281, 304, 328, 351,
    375, 398, 421, 445, 468, 492, 515, 539, 562, 585, 609, 632,
    656, 679, 703, 726, 750, 773, 796, 820, 843, 867, 890, 914,
    937, 960, 984, 1007, 1031, 1054, 1078, 1101, 1125, 1148, 1171,
    1195, 1218, 1242, 1265, 1289, 1312, 1335, 1359, 1382, 1406
)

seconds_to_KB_320 = (
    0, 195, 234, 273, 312, 351, 390, 429, 468, 507, 546, 585, 625,
    664, 703, 742, 781, 820, 859, 898, 937, 976, 1015, 1054, 1093,
    1132, 1171, 1210, 1250, 1289, 1328, 1367, 1406, 1445, 1484, 1523,
    1562, 1601, 1640, 1679, 1718, 1757, 1796, 1835, 1875, 1914, 1953,
    1992, 2031, 2070, 2109, 2148, 2187, 2226, 2265, 2304, 2343
)


class STATES():
    ANY = -1
    RESET = 0
    INIT = 1
    CONNECT = 2
    PLAY = 10
    TITLE = 11
    STOPPED = 12
    # Do not move it!
    PLAYER_ACTIVATED = 13

    CONNECT_ERROR = 100
    VOLUME = 101
    BUFF_MSG = 102
    BUFFER = 103

    ERROR_NO_PLAYER = 200
    ERROR_DEPENDENCY = 201
    ERROR_CONNECT = 202
    ERROR_START = 203

"""
Format of theme configuration
    Name, color_pair, foreground, background
If foreground == 0, color can be edited
If > 0, get color from list item referred to by number
Same for the background
"""
_param_to_color_id = {
    'Extra Func': (12, ),
    'PyRadio URL': (11, ),
    'Messages Border': (10, ),
    'Status Bar': (8, 9),
    'Stations': (1, 2),
    'Active Station': (3, ),
    'Active Cursor': (6, 7),
    'Normal Cursor': (4, 5),
}

THEME_ITEMS = (
    ('PyRadio URL', 2, 0, 3),
    ('Messages Border', 3, 0, 3),
    ('Status Bar', 7, 0, 0),
    ('Stations', 5, 0, 0),
    ('Active Station', 4, 0, 3),
    ('Normal Cursor', 6, 0, 0),
    ('Active Cursor', 9, 0, 0),
    ('Edit Cursor', 8, 0, 0)
)


def create_os_media_controller(identity="PyRadio", instance_name=None):
    """
    Factory: returns an OS-specific controller instance.
    Keep imports conditional to avoid cross-platform dependency leakage.
    """
    if platform.startswith("lin"):
        from .mpris import MprisController
        return MprisController(identity=identity, instance_name=instance_name)

    elif platform.startswith("win"):
        from .windows_smtc import WindowsSMTCController
        return WindowsSMTCController(identity=identity, instance_name=instance_name)

    elif platform.startswith("dar"):
        from .macos_media import MacOSMediaController
        return MacOSMediaController(identity=identity, instance_name=instance_name)

    return None

def describe_playlist(value):
    # Check if the value is within the range of the enum
    if value < 0 or value >= len(Station.__members__):
        # Return the default message if the value is out of range
        return f"Playlist has {Station.name.name} {Station.url.name}"

    # Collect all names from 0 to the given value
    names = [station.name for station in list(Station)[:value + 1]]

    # Join the names with spaces and return the result
    return 'Playlist has {}'.format(' '.join(names))

def remove_consecutive_empty_lines(input_list):
    cleaned_list = []
    previous_line_empty = False
    # remove empty fields
    work_list = [x for x in input_list if x]
    for line in work_list:
        stripped_line = line.strip()
        if stripped_line:
            cleaned_list.append(line)
            previous_line_empty = False
        elif not previous_line_empty:
            cleaned_list.append('\n')
            previous_line_empty = True
    return cleaned_list

def erase_curses_win(Y, X, beginY, beginX, char=' ', color=5):
    ''' empty a part of the screen
    '''
    empty_win = curses.newwin(
        Y - 2, X - 2,
        beginY + 1, beginX + 1
    )
    empty_win.bkgdset(char, curses.color_pair(color))
    empty_win.erase()
    empty_win.refresh()

def is_rasberrypi():
    ''' Try to detest rasberry pi '''
    try:
        with io.open('/sys/firmware/devicetree/base/model', 'r', encoding='utf-8') as m:
            if 'raspberry pi' in m.read().lower():
                return True
    except Exception:
        pass
    return False

    # if exists('/usr/bin/raspi-config'):
    #     return True
    # return False

def hex_to_rgb(hexadecimal):
    n = hexadecimal.lstrip('#')
    return tuple(int(n[i:i+2], 16) for i in (0, 2, 4))

def rgb_to_hex(rgb):
    return '#%02x%02x%02x' % rgb

def curses_rgb_to_hex(rgb):
    return rgb_to_hex(tuple(int(y * 255 / 1000) for y in rgb))

def rgb_to_curses_rgb(rgb):
    return tuple(int(y *1000 / 255) for y in rgb)


class RichColorPrinter:
    """
    A conditional printer that handles rich text formatting tags.

    Provides colorized output when enabled, while preserving non-rich bracket
    notation (like [X] markers) when color is disabled.

    Args:
        use_color (bool): If True, uses rich formatting; if False, removes rich tags

    Attributes:
        use_color (bool): Color printing state
        _rich_print (callable): Cached rich print function

    Methods:
        __call__(message): Print message with conditional formatting

    Initialization:
        >>> printer = RichColorPrinter(use_color=not no_color)

    Usage:
        >>> printer = RichColorPrinter(use_color=True)
        >>> printer("[green]Success![/green]")  # Colorized
        >>> printer = RichColorPrinter(use_color=False)
        >>> printer("[green]Success![/green]")  # Plain text: "Success!"

    Note:
        Only removes rich formatting tags (colors, styles), preserves other
        bracket content like [X] markers or [custom_text].
    """

    def __init__(self, use_color=True):
        self.use_color = use_color
        self._rich_print = None
        # Define specific rich tags to remove (colors and styles only)
        self.rich_tags = [
            'red', 'green', 'blue', 'yellow', 'cyan', 'magenta', 'white', 'black',
            'bold', 'italic', 'underline', 'dim', 'reverse', 'blink', 'strike',
            '/red', '/green', '/blue', '/yellow', '/cyan', '/magenta', '/white', '/black',
            '/bold', '/italic', '/underline', '/dim', '/reverse', '/blink', '/strike'
        ]

    def __call__(self, message):
        """
        Print message with conditional formatting handling.

        Args:
            message (str): Message to print, may contain rich formatting tags
        """
        if not self.use_color:
            import re
            pattern = r'\[(' + '|'.join(self.rich_tags) + r')\]'
            clean_message = re.sub(pattern, '', message)
            print(clean_message)
        else:
            if self._rich_print is None:
                from rich import print as rich_print
                self._rich_print = rich_print
            self._rich_print(message)


class StationsChanges():
    '''
    #########################################################################
    #                                                                       #
    #                         stations.csv change                           #
    #                                                                       #
    #########################################################################
    #                                                                       #
    # This section will include the changes of stations.csv                 #
    #                                                                       #
    # version_changed                                                       #
    #   The version the changes appeared (calculated)                       #
    #                                                                       #
    # It will also contain three lists:                                     #
    #   added     - list of stations added                                  #
    #   changed   - list of stations changed                                #
    #   deleted   - list of stations deleted                                #
    #                                                                       #
    # The "changed" list will be of format:                                 #
    #           [[ x , [station name, station url ]]                        #
    # where:                                                                #
    #           x : 0 / 1 index changed (0: name, 1: URL)                   #
    #                                                                       #
    #########################################################################
    '''

    def __init__(self, config):
        '''
        versions = {
            (0, 9, 2):                       # 0.9.2 version
            [
                [...........],                 # added
                [x, [...........]],            # changed
                [...........],                 # deleted
            ],
            (0, 9, 1):                       # 0.9.1 version
            [
                [...........],                 # added
                [x, [...........]],            # changed
                [...........],                 # deleted
            ]
        ]
        '''
        self.versions = {
            (0, 9, 2):
            [
                [
                    ['Groove Salad Classic (Early 2000s Ambient)', 'https://somafm.com/gsclassic.pls'],
                    ['n5MD Radio (Ambient and Experimental)', 'https://somafm.com/n5md.pls'],
                    ['Vaporwaves [SomaFM]', 'https://somafm.com/vaporwaves.pls'],
                    ['The Trip: [SomaFM]', 'https://somafm.com/thetrip.pls'],
                    ['Heavyweight Reggae', 'https://somafm.com/reggae.pls'],
                    ['Metal Detector', 'https://somafm.com/metal.pls'],
                    ['Synphaera Radio (Space Music)', 'https://somafm.com/synphaera.pls']
                ], # added

                [
                    [0, ['Reggae Dancehall (Ragga Kings)', 'https://raggakings.radio:8443/stream.ogg']]
                ], # changed

                [] # deleted
            ],

            (0, 9, 3):
            [
                [
                    ['Radio Levač (Serbian Folk & Country)', 'http://213.239.205.210:8046/stream'],
                    ['Radio 35 (Serbian and English Pop, Folk, Country & Hits)', 'http://stream.radio035.net:8010/listen.pls']
                ], # added
                [], # changed
                [], # deleted
            ],

            (0, 9, 3, 11, 5):
            [
                [], # added
                [
                    ['DanceUK', 'http://uk2.internet-radio.com:8024/listen.pls'],
                    ['JazzGroove', r'http://199.180.72.2:8015/listen.pls\?sid\=1'],
                    ['Metal Detector' 'http://somafm.com/metal.pls'],
                ], # changed
                [
                    ['Beyond Metal (Progressive - Symphonic)', 'http://streamingV2.shoutcast.com/BeyondMetal'],
                    ['Vox Noctem: Rock-Goth', 'http://r2d2.voxnoctem.de:8000/voxnoctem.mp3'],
                ], # deleted
            ],
        }
        self.version_changed = None
        self.keys = None
        self._stations = None
        self._stations_file = None
        self._playlist_version = 0
        self._cnf = config
        self._last_sync_file = join(self._cnf.state_dir, 'last-sync')
        self._asked_sync_file = join(self._cnf.state_dir, 'asked-sync')
        self.PLAYLIST_HAS_NAME_URL = 0
        self.PLAYLIST_HAS_NAME_URL_ENCODING = 1
        self.PLAYLIST_HAS_NAME_URL_ENCODING_ICON = 2
        self.counts = [0, 0, 0]
        self._out_stations_file = None
        self._bck_stations_file = None
        self.asked_sync = None
        self.last_sync = None
        self.version_to_write = None

    def _read_synced_version(self, asked=False):
        in_file = self._asked_sync_file if asked else self._last_sync_file
        # print('in_file = "{}"'.format(in_file))
        if exists(in_file):
            try:
                with open(in_file, 'r', encoding='utf-8') as sync_file:
                    line = sync_file.readline().strip()
                    try:
                        line = tuple([int(ver.strip()) for ver in line.split(',')])
                    except (ValueError, AttributeError):
                        return None
                    return line
            except:
                pass
        return None

    def write_synced_version(self, asked=False):
        out_file = self._asked_sync_file if asked else self._last_sync_file
        try:
            with open(out_file, 'w', encoding='utf-8') as sync_file:
                sync_file.write(self.version_to_write)
        except:
            return -5 if asked else -6
        return -3 if asked else 0

    def _open_stations_file(self):
        self._stations = []
        self._stations_file = join(self._cnf.stations_dir, 'stations.csv')
        self._playlist_version = self.PLAYLIST_HAS_NAME_URL
        if exists(self._stations_file):
            with open(self._stations_file, 'r', encoding='utf-8') as cfgfile:
                try:
                    for row in csv.reader(filter(lambda row: row[0]!='#', cfgfile), skipinitialspace=True):
                        if not row:
                            continue
                        try:
                            name, url = [s.strip() for s in row]
                            self._stations.append([name, url, '', ''])
                        except:
                            try:
                                name, url, enc = [s.strip() for s in row]
                                self._stations.append([name, url, enc, ''])
                                self._playlist_version = self.PLAYLIST_HAS_NAME_URL_ENCODING
                            except:
                                name, url, enc, onl = [s.strip() for s in row]
                                self._stations.append([name, url, enc, onl])
                                self._playlist_version = self.PLAYLIST_HAS_NAME_URL_ENCODING_ICON
                except:
                    self._stations = []
                    self._playlist_version = self.PLAYLIST_HAS_NAME_URL
                    return False
            return True
        return False

    def _save_stations_file(self, print_messages=True):
        self._out_stations_file = join(self._cnf.stations_dir, 'stations-new.csv')
        self._bck_stations_file = join(self._cnf.stations_dir, 'stations.csv.bck')
        try:
            with open(self._out_stations_file, 'w', encoding='utf-8') as cfgfile:
                writter = csv.writer(cfgfile)
                for a_station in self._stations:
                    if a_station[3] != '':
                        a_station[3] = a_station[3]
                    writter.writerow(self._format_playlist_row_out(a_station))
        except:
            print('Error: Cannot create the updated stations file.')
            print('       The updated stations file would be\n         "{}".'.format(self._out_stations_file))
            return False
        ''' rename stations.csv to stations.csv.bck '''
        try:
            rename(self._stations_file, self._bck_stations_file)
        except:
            print('Error: Cannot create the stations backup file.')
            print(f'       The updated stations file can be found at\n         "{self._out_stations_file}".')
            return False
        ''' rename stations-new.csv to stations.csv '''
        try:
            rename(self._out_stations_file, self._stations_file)
        except:
            print('Error: Cannot rename the updated stations file.')
            print(f'       The updated stations file can be found at\n         "{self._out_stations_file}".')
            print(f'       The old stations file has been backed up as\n         "{self._bck_stations_file}".')
            return False
        ''' remove bck file '''
        try:
            remove(self._bck_stations_file)
        except:
            pass
        if print_messages:
            print('File "stations.csv" updated...')
        return True

    def _format_playlist_row_out(self, a_row):
        ''' Return a 2-column if in old format,
            a 3-column row if has encoding, or
            a 4 column row if has online browser flag too '''
        if self._playlist_version == self.PLAYLIST_HAS_NAME_URL_ENCODING_ICON:
            return a_row
        if self._playlist_version == self.PLAYLIST_HAS_NAME_URL_ENCODING:
            return a_row[:-1]
        return a_row[:-2]

    def _format_vesion(self, a_version_tuple):
        ret = str(a_version_tuple)
        ret = ret.replace('(', '')
        ret = ret.replace(')', '')
        ret = ret.replace(', ', '.')
        return ret

    def check_if_version_needs_sync(self, stop=None):
        ''' check if we need to sync stations.csv
            takes under consideration the answer
            the user gave at the TUI
        '''
        ret = self.stations_csv_needs_sync(print_messages=False)
        if stop is not None:
            if stop():
                if logger.isEnabledFor(logging.DEBUG):
                    logger.debug('asked to stop! Terminating!')
                return False
        self.asked_sync = self._read_synced_version(asked=True)
        if stop is not None:
            if stop():
                if logger.isEnabledFor(logging.DEBUG):
                    logger.debug('asked to stop! Terminating!')
                return False
        if self.version_changed == self.asked_sync:
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug('asked_sync is equal to version_changed!!!')
            return False
        return ret

    def stations_csv_needs_sync(self, print_messages=True, stop=None):
        ''' check if we need to sync stations.csv
            it will return true no matter what the user has
            replied about syncing, at the TUI

            Used by update_stations_csv()
        '''
        self.keys = list(self.versions)
        self.keys.sort()
        # print('keys = {}'.format(self.keys))
        if stop is not None:
            if stop():
                if logger.isEnabledFor(logging.DEBUG):
                    logger.debug('asked to stop! Terminating!')
                return False
        self.last_sync = self._read_synced_version()
        if stop is not None:
            if stop():
                if logger.isEnabledFor(logging.DEBUG):
                    logger.debug('asked to stop! Terminating!')
                return False
        if exists(self._last_sync_file):
            try:
                with open(self._last_sync_file, 'r', encoding='utf-8') as sync_file:
                    line = sync_file.readline().strip()
                    self.last_sync = line
            except:
                ret = False
            if self.last_sync is None:
                ret = True
            else:
                try:
                    last_sync = tuple(int(ver.strip()) for ver in self.last_sync.split(','))
                except (ValueError, AttributeError):
                    # make sure ret gets False
                    last_sync = self.keys[-1]
                ret = self.keys[-1] > last_sync
        else:
            if stop is not None:
                if stop():
                    if logger.isEnabledFor(logging.DEBUG):
                        logger.debug('asked to stop! Terminating!')
                    return False
            ret = True

        if ret and self.last_sync is not None:
            logger.error(f'==\n==\n{self.last_sync = }\n==\n==')
            self.keys.reverse()
            logger.error(f'==\n==\n{self.keys = }\n==\n==')
            logger.error(f'==\n==\n{self.keys[-1] = }\n==\n==')

            if isinstance(self.last_sync, str):
                try:
                    last_sync = tuple(int(n.strip()) for n in self.last_sync.split(','))
                except (ValueError, AttributeError) as e:
                    logger.error(f'last_sync exception: {e}')
                    last_sync = self.keys[-1]
            else:
                last_sync = self.last_sync
            while self.keys[-1] <= last_sync:
                self.keys.pop()
            self.keys.reverse()
        if stop is not None:
            if stop():
                if logger.isEnabledFor(logging.DEBUG):
                    logger.debug('asked to stop! Terminating!')
                return False
        # print('keys = {}'.format(self.keys))
        self.version_changed = self.keys[-1]
        self.version_to_write = str(self.version_changed).replace('(', '').replace(')', '')
        if stop is not None:
            if stop():
                if logger.isEnabledFor(logging.DEBUG):
                    logger.debug('asked to stop! Terminating!')
                return False
        if print_messages:
            print('Updating "stations.csv"')
            print('Last updated version: {}'.format(self._format_vesion(self.version_changed)))
            print(' Last synced version: {}'.format(self._format_vesion(self.last_sync)))
        if stop is not None:
            if stop():
                if logger.isEnabledFor(logging.DEBUG):
                    logger.debug('asked to stop! Terminating!')
                return False
        if print_messages and not ret:
            print('Already synced: "stations.csv"')
        return ret

    def _format_playlist_row_in(self, a_row):
        ''' Return a 2-column if in old format,
            a 3-column row if has encoding, or
            a 4 column row if has online browser flag too '''
        while len(a_row) < 4:
            a_row.append('')
        return a_row

    def update_stations_csv(self, print_messages=True):
        ''' update stations.csv
            Returns:
                 1 : Update not needed
                 0 : All ok
                -1 : Cannot read stations.csv
                -2 : File not saved


        '''
        # if self.stations_csv_needs_sync(print_messages=print_messages):
        if self.stations_csv_needs_sync(print_messages=False):
            if not self._open_stations_file():
                if print_messages:
                    print('Cannot read "stations.csv"')
                return -1
            # for n in self._stations:
            #     print(n)

            for k in self.keys:
                if print_messages:
                    print('  From version: {}'.format('.'.join(map(str, k))))
                for n in self.versions[k][2]:
                    found = [x for x in self._stations if x[0] == n[0]]
                    if found:
                        for an_item in found:
                            if print_messages:
                                print('[red]    --- deleting: "[green]{}[/green]"[/red]'.format(an_item[0]))
                            self.counts[2] += 1
                            self._stations.pop(self._stations.index(an_item))
                for n in self.versions[k][1]:
                    found = []
                    if n[0] == 0:
                        found = [x for x in self._stations if x[0] == n[1][0] and x[1] != n[1][1]]
                    elif n[0] == 1:
                        found = [x for x in self._stations if x[1] == n[1][1] and x[0] != n[1][0]]
                    if found:
                        if print_messages:
                            print('[plum4]    +/- updating: "[green]{}[/green]"[/plum4]'.format(found[0][0]))
                        self.counts[1] += 1
                        index = self._stations.index(found[0])
                        self._stations[index] = self._format_playlist_row_in(n[1])
                for n in self.versions[k][0]:
                    found = [x for x in self._stations if x[0] == n[0]]
                    if not found:
                        if print_messages:
                            print('[magenta]    +++   adding: "[green]{}[/green]"[/magenta]'.format(n[0]))
                        self.counts[0] += 1
                        self._stations.append(self._format_playlist_row_in(n))

            if self._save_stations_file(print_messages=print_messages):
                ret = self.write_synced_version()
                if ret == -6:
                    if print_messages:
                        txt = '''
[red]Error:[/red] [magenta]PyRadio[/magenta] could not write the "last_sync" file.
This means that although stations have been synced, [magenta]PyRadio[/magenta] will try
to sync them again next time, which means that you may end up with
duplicate stations.

Please close all open programs and documents and create the file
[green]{0}[/green]
and write in it
      "[green]{1}[/green]" (no quotes).
                        '''.format(
                            self._last_sync_file,
                            self.version_to_write
                        )
                        print(txt)

                elif print_messages:
                    print('\n[bold]Summary[/bold]\n[magenta]    +++ added   :[/magenta]  {0}\n[plum4]    +/- updated :[/plum4]  {1}\n[red]    --- deleted :[/red]  {2}'.format(self.counts[0], self.counts[1], self.counts[2]))
                return ret
            return -2
        return 1

        # print('\n\n\n')
        # for n in self._stations:
        #     print(n)


def install_dbus_next_or_fast(the_module, dbus_next_file):
    import subprocess
    def run(cmd):
        try:
            p = subprocess.run(
                cmd,
                capture_output=True,
                text=True,
                check=False,
            )
            return p.returncode, p.stdout or ''
        except Exception:
            return 1, ''

    def check_dbus_next_file(the_module, dbus_next_file):
        try:
            with open(dbus_next_file, 'w') as f:
                f.write(the_module)
        except OSError as e:
            file_dir = basename(dbus_next_file)
            if e.errno == errno.EACCES:
                print(f'[bold red]Error: [/bold red]Cannot write in {file_dir}.')
            elif e.errno == errno.ENOENT:
                print(f'[bold red]Error: [/bold red]The directory does not exit: {file_dir}.')
            elif e.errno == errno.ROFS:
                print('[bold red]Error: [/bold red]This filesystem is readonly.')
            else:
                print(f'[bold red]Error: [/bold red]Cannot create file: {e.strerror}')
            return 1
        except Exception as e:
            print(f'[bold red]Error: [/bold red]{e}')
            return 1
        return 0

    # find pipx
    pipx_base = None
    for candidate in (
        ['pipx'],
        ['python', '-m', 'pipx'],
        ['python3', '-m', 'pipx'],
    ):
        rc, _ = run(candidate + ['--version'])
        if rc == 0:
            pipx_base = candidate
            break

    if not pipx_base:
        print('[bold magenta]PyRadio[/bold magenta] is not in an isolated environment')
        return 1

    rc, output = run(pipx_base + ['list', '--include-injected'])
    if rc != 0:
        print('[bold magenta]PyRadio[/bold magenta] is not in an isolated environment')
        return 1

    if 'pyradio' not in output:
        print('[bold magenta]PyRadio[/bold magenta] not installed through pipx')
        return 1

    # must have at least one injected package
    # psutil is always present in an isolated environment
    if 'psutil' not in output:
        print('[bold magenta]PyRadio[/bold magenta] is not in an isolated environment')
        return 1

    # if already injected
    if f'dbus-{the_module}' in output:
        print(f'dbus-{the_module} injected successfully')
        return check_dbus_next_file(the_module, dbus_next_file)

    # perform injection
    rc, _ = run(pipx_base + ['inject', 'pyradio', f'dbus-{the_module}'])
    if rc == 0:
        print(f'dbus-{the_module} injected successfully')
        return check_dbus_next_file(the_module, dbus_next_file)
    else:
        print('[bold magenta]PyRadio[/bold magenta] is not in an isolated environment')
        return 1
    return 0


class CsvReadWrite():
    ''' A base class to read and write a PyRadio playlist '''

    def __init__(self, a_file=None, encoding_to_remove=None):
        self._items = None
        self._version = Station.url
        self.encoding_to_remove = encoding_to_remove
        # We save the original file as given
        self._original_file = a_file
        # For internal usage, we save a version that can be used by other functions
        self._file = self._normalize_file_path(a_file) if a_file else None

    @property
    def items(self):
        return self._items

    @items.setter
    def items(self, value):
        self._items = value[:]

    @property
    def version(self):
        return self._version

    @version.setter
    def version(self, value):
        self._version = value

    @property
    def groups(self):
        if self._items:
            return [i for i, x in enumerate(self._items) if x[1] == '-']
        return None

    def _normalize_file_path(self, file_path):
        """Convert file_path to a format that can be used in all cases."""
        if file_path is None:
            return None
        # If it is a Traversable, return it as it is
        if isinstance(file_path, Traversable):
            return file_path
        # If it is str, convert to Path
        if isinstance(file_path, str):
            return Path(file_path)
        # If it is Path, return it as it is
        if isinstance(file_path, Path):
            return file_path
        raise TypeError(f"Unsupported file type: {type(file_path)}")

    def _get_writable_file(self, file_path):
        """Return a string path for writing."""
        if file_path is None:
            return None
        # We always need a string path
        if isinstance(file_path, (str, Path)):
            return str(file_path)
        raise TypeError(f"Cannot write to {type(file_path)}. Must be str or Path")

    def read(self, a_file=None):
        ''' Read a PyRadio playlist

            The file is a_file or self._file (if a_file is None)
            Supports str, Path, and Traversable objects
            Populates self._items and self._version
            Returns True or False (if error)
        '''
        current_version = Station.url
        # Use a_file if given, otherwise use self._file
        in_file = a_file if a_file is not None else self._file

        if in_file is None:
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug('No file specified for reading')
            return False

        # Path normalization
        in_file = self._normalize_file_path(in_file)
        self._items = []

        try:
            # We handle Traversable differently from the others
            if isinstance(in_file, Traversable):
                # We open Traversable with its own open() method
                with in_file.open(encoding='utf-8') as cfgfile:
                    return self._read_csv(cfgfile, current_version)
            # We open a file with the plain open()
            with open(str(in_file), 'r', encoding='utf-8') as cfgfile:
                return self._read_csv(cfgfile, current_version)

        except (FileNotFoundError, IOError, OSError) as e:
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug(f'Cannot open playlist file: {e}')
            self._items = []
            self._version = current_version
            return False

    def _read_csv(self, file_handle, current_version):
        """Internal function to read a CSV from a file handle."""
        try:
            for row in csv.reader(filter(lambda row: row[0] != '#', file_handle), skipinitialspace=True):
                if not row:
                    continue

                # Initialize variables with default values
                name = url = enc = icon = volume = http = referer = profile = buffering = player = ''
                this_row_version = Station.url
                # Assign values based on the length of the row
                row_length = len(row)
                name = row[0].strip()
                url = row[1].strip()
                if row_length > Station.encoding:
                    enc = row[Station.encoding].strip()
                    this_row_version = Station.encoding
                if row_length > Station.icon:
                    icon = row[Station.icon].strip()
                    this_row_version = Station.icon
                if row_length > Station.profile:
                    profile = row[Station.profile].strip()
                    this_row_version = Station.profile
                if row_length > Station.buffering:
                    buffering = row[Station.buffering].strip()
                    this_row_version = Station.buffering
                if row_length > Station.volume:
                    volume = row[Station.volume].strip()
                    this_row_version = Station.volume
                if row_length > Station.http:
                    http = row[Station.http].strip()
                    this_row_version = Station.http
                if row_length > Station.referer:
                    referer = row[Station.referer].strip()
                    this_row_version = Station.referer
                if row_length > Station.player:
                    player = row[Station.player].strip()
                    this_row_version = Station.player

                if buffering:
                    if '@' not in buffering:
                        buffering += '@128'
                else:
                    buffering = '0@128'

                if self.encoding_to_remove is not None:
                    if enc == self.encoding_to_remove:
                        enc = ''

                # Append the parsed values to the reading stations list
                station_info = [
                    name, url, enc, icon if icon else '',
                    profile, buffering, http, volume, referer, player
                ]
                self._items.append(station_info)

                # Update playlist version based on the presence of optional fields
                current_version = max(this_row_version, current_version)
                self._version = current_version

            return True

        except (csv.Error, ValueError) as e:
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug(f'Playlist is malformed: {e}')
            self._items = []
            self._version = current_version
            return False

    def _format_playlist_row(self, a_row):
        ''' Return a formatted row (list)
            Functionality:
                Removes {'image': '...'}
                Removes '0@128'
                Eliminates any trailing empty fields
        '''
        this_row = a_row[:]
        if self.encoding_to_remove is not None:
            if this_row[Station.encoding] == self.encoding_to_remove:
                this_row[Station.encoding] = ''

        if len(this_row) > Station.buffering:
            if this_row[Station.buffering] == '0@128':
                this_row[Station.buffering] = ''
        while this_row and this_row[-1] == '':
            this_row.pop()
        return this_row

    def write(self, a_file=None, items=None):
        ''' Save a PyRadio playlist
        Creates a txt file and write stations in it.
        Then renames it to final target

        Supports str and Path objects for writing

        Returns   0: All ok
                 -1: Error writing file
                 -2: Error renaming file
        '''
        # Use a_file if given, otherwise use self._file
        out_file = a_file if a_file is not None else self._file
        out_items = items if items else self._items

        if out_file is None:
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug('No file specified for writing')
            return -1

        # For writing, we need a string path
        try:
            out_file_str = self._get_writable_file(out_file)
        except TypeError as e:
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug(f'Unsupported file type for writing: {e}')
            return -1

        txt_out_file = out_file_str.replace('.csv', '.txt')

        try:
            with open(txt_out_file, 'w', encoding='utf-8') as cfgfile:
                writer = csv.writer(cfgfile)
                writer.writerow(['# PyRadio Playlist File Format:'])
                writer.writerow(
                    ['# name', 'url', 'encoding', 'icon',
                     'profile', 'buffering', 'force-http',
                     'volume', 'referer', 'player'])
                for a_station in out_items:
                    writer.writerow(self._format_playlist_row(a_station))
        except (IOError, OSError, UnicodeEncodeError) as e:
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug(f'Cannot open playlist file for writing: {e}')
            return -1

        try:
            move(txt_out_file, out_file_str)
        except (shutil_Error, FileNotFoundError, PermissionError) as e:
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug(f'Cannot rename playlist file: {e}...')
            return -2
        return 0


class ProfileManager():

    def __init__(self):
        self._config_file = None

    @property
    def config_files(self):
        if self._config_file is None:
            self._config_files = self.get_config_files()
        return self._config_files

    @config_files.setter
    def config_files(self, value):
        self._config_files = value

    def reread_files(self):
        self._config_file = None
        return self.config_files

    @classmethod
    def get_config_files(cls):
        out = {}
        ''' MPV config files '''
        if platform.startswith('win'):
            config_files = [join(getenv('APPDATA'), "mpv", "mpv.conf")]
        else:
            # linux, freebsd, etc.
            xdg_config = getenv('XDG_CONFIG_HOME')
            if xdg_config:
                config_files = [xdg_config + "/mpv/mpv.conf"]
            else:
                config_files = [expanduser("~") + "/.config/mpv/mpv.conf"]
            config_files.append("/etc/mpv/mpv.conf")
            config_files.append("/usr/local/etc/mpv/mpv.conf")
        out['mpv'] = config_files[:]

        ''' MPlayer config files '''
        config_files = [expanduser("~") + "/.mplayer/config"]
        if platform.startswith('win'):
            if exists(r'C:\\mplayer\\mplayer.exe'):
                config_files[0] = r'C:\\mplayer\mplayer\\config'
            elif exists(join(getenv('USERPROFILE'), "mplayer", "mplayer.exe")):
                config_files[0] = join(getenv('USERPROFILE'), "mplayer", "mplayer", "config")
            elif exists(join(getenv('APPDATA'), "pyradio", "mplayer", "mplayer.exe")):
                config_files[0] = join(getenv('APPDATA'), "pyradio", "mplayer", "mplayer", "config")
            else:
                config_files = []
        else:
            config_files.append("/usr/local/etc/mplayer/mplayer.conf")
            config_files.append('/etc/mplayer/config')
        out['mplayer'] = config_files[:]
        # logger.error(f'{out =  }')
        return out

    def set_vlc_config_file(self, config):
        self._config_files['vlc'] = config

    def all_profiles(self):
        # out = {'mpv': [], 'mplayer': [], 'vlc': []}
        out = {'mpv': [], 'mplayer': []}
        try:
            # for player_name in ('mpv', 'mplayer', 'vlc'):
            for player_name in ('mpv', 'mplayer'):
                profiles = self.profiles(player_name)
                # logger.error('{}: {}'.format(player_name, profiles))
                if profiles:
                    out[player_name] = profiles[:]
        except KeyError:
            pass
        result = self._create_profile_list(out)
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug(f'all_profiles: {result = }')
        return result

    def _create_profile_list(self, players_dict):
        # Create a set of profile names for each player
        profile_sets = [set(profiles) for profiles in players_dict.values()]

        # Find the intersection of all profile sets
        common_profiles = set.intersection(*profile_sets)

        # Initialize the result list with common profiles
        result = list(common_profiles)

        # Add profiles that are not common, appending the player name
        for player, profiles in players_dict.items():
            for profile in profiles:
                if profile not in common_profiles:
                    result.append(f'{player}: {profile}')

        return sorted(result)

    def profiles(self, player_name):
        profiles_list = []
        for a_file in self.config_files[player_name]:
            config_list = self._read_a_config_file(a_file)
            if config_list:
                config_list = [ x.strip() for x in config_list ]
                profiles = [ x[1:-1] for x in config_list if x.startswith('[') and x.endswith(']') ]
                if profiles:
                    profiles_list.extend(profiles)
        profiles_list = list(set(profiles_list))
        profiles_list = [ x for x in profiles_list if x not in ('pyradio-volume','silent')]
        return sorted(profiles_list)

    def _read_a_config_file(self, config_file):
        # Read config file
        if not exists(config_file):
            return []
        try:
            with open(config_file, 'r', encoding='utf-8') as file:
                ret = [line.lstrip() if line.strip() else line for line in file.readlines()]
        except Exception:
            ret = []
        return ret

    def _write_config_file(self, config_file, a_string=None, a_list=None):
        try:
            makedirs(dirname(config_file), exist_ok=True)
            with open(config_file, 'w', encoding='utf-8') as file:
                if a_string is None:
                    file.writelines(a_list)
                else:
                    file.write(a_string)
            return True
        except (FileNotFoundError, PermissionError, IsADirectoryError,
                UnicodeEncodeError, OSError, IOError):
            if logger.isEnabledFor(logging.DEBUG):
                if a_list:
                    a_string='\n'.join(a_list)
                logger.debug('Error writing profile string\n{}\nin "{}"'.format(a_string, config_file))
            return False

    def save_volume(self, player_name, profile_name, volume):
        ''' save volume value in a profile

            Parameters
                player_name         : string
                profile_name        : string
                volume              : any

            Returns
                profile_name        : if success
                None                : if failure
        '''
        # logger.error('before profile name: "{}"'.format(profile_name))
        config_file = self.config_files[player_name][0]
        if profile_name.startswith('[') and \
                profile_name.endswith(']'):
            profile_name = profile_name[1:-1]
        # logger.error('after  profile name: "{}"'.format(profile_name))
        if not exists(config_file):
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug(
                    '[{}] file not found! Adding profile [{}] with volume {}\n\n'.format(
                        config_file, profile_name, volume
                    )
                )
            makedirs(dirname(config_file), exist_ok=True)
            return self.append_to_config(player_name, profile_name, 'volume=' + str(volume))

        config_list = self._read_a_config_file(config_file)
        config_string = ''.join(config_list)
        # logger.error(f'{config_file = }')
        # logger.error(f'before remove {config_list = }')
        if any(line.strip() == '[' + profile_name + ']' for line in config_list):
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug('[{}] profile found!\n\n'.format(profile_name))
            config_list = remove_consecutive_empty_lines(config_string.split('\n'))
            config_string = '\n'.join(config_list)
            # logger.error(f'after  remove {config_string = }')
            # find all lines starting with '['
            indexes = [i for i, x  in enumerate(config_list) if x.startswith('[')]
            # logger.error(f'{indexes = }')
            # get [profile_name] index
            profile_index = config_list.index('[' + profile_name + ']')
            # logger.error(f'{profile_index = }')
            volume_adjusted = False
            for i in range(profile_index+1, len(config_list)):
                # logger.error('checking "{}"'.format(config_list[i]))
                if config_list[i].startswith('volume='):
                    config_list[i] = 'volume=' + str(volume)
                    volume_adjusted = True
                    break
                if i in indexes:
                    # volume not found in profile
                    break
            if not volume_adjusted:
                config_list[profile_index] += '\nvolume={}\n'.format(volume)
            config_list = ['\n' + x if x.startswith('[') else x for x in config_list]
            # logger.error('\n\n')
            # for n in config_list:
            #     logger.error(f'"{n}"')
            # logger.error(config_list)
            # logger.error('\n'.join(config_list))
            # logger.error('\n\n')
            ret = self._write_config_file(config_file, a_string='\n'.join(config_list) + '\n\n')
            if ret:
                return profile_name
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug('Error writing profile [{}] in "{}"'.format(profile_name, config_file))
            return None

        if logger.isEnabledFor(logging.DEBUG):
            logger.debug('[{}] profile not found! Adding it with volume'.format(profile_name))
        return self.append_to_config(player_name, profile_name, 'volume=' + str(volume))

    def write_silenced_profile(self, player_name):
        self.add_to_config(player_name, 'silent', 'volume=0')

    def add_to_config(self, player_name, profile_name, profile_string):
        ''' appends a profile to the user's player config file
            it is a wrapper function for append_to_config

            checks if profile already exists in user's config file

            Paramaters
                player_name         : string
                profile_name        : string
                profile_contents    : string

            Returns
                profile_name        : if success or profile exists
                None                : if failure

        '''
        for config_file in self.config_files[player_name]:
            if exists(config_file):
                config_string = self._read_a_config_file(config_file)
                if '[' + profile_name + ']' in config_string:
                    if logger.isEnabledFor(logging.DEBUG):
                        logger.debug('[{}] profile found!'.format(profile_name))
                    return profile_name

        ''' profile not found in config
            create a default profile
        '''
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug(f'No [{profile_name}] profile found!')
        # fix for #229
        base_path = dirname(self.config_files[player_name][0])
        if not exists(base_path):
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug(f'Dir created: "{base_path}')
            makedirs(base_path)
        ret = self.append_to_config(player_name, profile_name, profile_string)
        if ret is None:
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug('Cannot wirte [{}] profile in: "{}"'.format(profile_name, self.config_files[player_name][0]))
        return ret

    def append_to_config(self, player_name, profile_name, profile_contents):
        ''' appends a profile to the user's player config file

            Paramaters
                player_name         : string
                profile_name        : string
                profile_contents    : string

            Returns
                profile_name        : if success
                None                : if failure
        '''
        try:
            config_file = self.config_files[player_name][0]

            # Read the current contents of the config file
            config_list = self._read_a_config_file(config_file)

            # append new profile
            if profile_name and profile_contents:
                config_list.append('[' + profile_name + ']\n')
                config_list.append(profile_contents + '\n\n')

            # Remove consecutive empty lines, leaving at least one
            cleaned_list = remove_consecutive_empty_lines(config_list)

            # Write back the updated list to the config file
            self._write_config_file(config_file, a_list=cleaned_list)
            profile = profile_name
        except (FileNotFoundError, PermissionError, IsADirectoryError,
                UnicodeEncodeError, OSError, IOError):
            profile = None
        return profile

    def copy_profile_with_new_volume(self, player_name, existing_profile, new_profile, volume):
        ''' copy a profile and chnge the volume in the new profile

            Parameters:
                player_name         : the player
                existing_profile    : the source profile
                new_profile         : the new profile
                volume              : volume value

            if new_profile exists, it will be overwritten

            Returns:
                new_profile         : success
                None                : failed to write the file
        '''
        if existing_profile:
            existing_profile = existing_profile.replace('[', '').replace(']', '')
        new_profile = new_profile.replace('[', '').replace(']', '')

        # Read config file
        out_file = self.config_files[player_name][0]
        config_file = self.config_files[player_name][0]
        config_list = self._read_a_config_file(config_file)

        # Extract existing profile section if it exists
        existing_section = None
        if existing_profile:
            # Find existing profile section
            section = []
            in_section = False
            for line in config_list:
                stripped = line.strip()
                if stripped == f'[{existing_profile}]':
                    in_section = True
                    section.append(line)
                elif in_section:
                    if stripped.startswith('['):
                        break
                    section.append(line)
            if in_section:
                existing_section = section

        # Create or update the new profile section
        if existing_section:
            # Update volume in copied section
            new_section = []
            volume_found = False
            for line in existing_section:
                stripped = line.strip()
                if stripped.startswith('volume='):
                    new_section.append(f'volume={volume}\n')
                    volume_found = True
                else:
                    new_section.append(line)
            if not volume_found:
                new_section.append(f'volume={volume}\n')
            # Rename profile
            new_section[0] = f'[{new_profile}]\n'
        else:
            # Create new section
            new_section = [f'[{new_profile}]\n', f'volume={volume}\n']

        # Remove existing new_profile entries
        cleaned_config = []
        skip = False
        for line in config_list:
            stripped = line.strip()
            if stripped == f'[{new_profile}]':
                skip = True
            elif skip:
                if stripped.startswith('['):
                    skip = False
                    cleaned_config.append(line)
            else:
                cleaned_config.append(line)

        # Add new profile section
        cleaned_config += ['\n'] + new_section

        # Clean up formatting
        # cleaned_config = ['\n' + x if x.startswith('[') else x for x in cleaned_config]
        cleaned_config = remove_consecutive_empty_lines(cleaned_config)

        # Write back to file
        ret = self._write_config_file(out_file, a_list=cleaned_config)
        if ret:
            return new_profile
        logger.error(f"Error saving {player_name} config")
        return None


def validate_resource_opener_path(a_file):
    # Check if the file exists
    if not exists(a_file):
        # If the file doesn't exist, try to find it using shutil.which
        full_path = which(a_file)
        if full_path is None:
            return None
        a_file = full_path
    # Check if the file is executable
    if not access(a_file, X_OK):
        return None
    # Return the validated path
    return a_file

def get_cached_icon_path(cache_dir, station_name, icon_url):
    """
    Get cached icon path, migrating from station-based to URL-based naming
    """
    # Old filename (based on station name)
    old_filename = f"{station_name.replace(' ', '_')}."
    ext = 'jpg' if icon_url.endswith(('.jpg', '.jpeg')) else 'png'
    path1 = join(cache_dir, old_filename + ext)

    # New filename (based on icon URL hash)
    url_hash = hashlib.md5(icon_url.encode()).hexdigest()
    new_filename = f"{url_hash}.{ext}"
    path2 = join(cache_dir, new_filename)

    # Migration logic
    if exists(path1):
        if exists(path2):
            # Both new and old filenames exist - delete old
            try:
                remove(path1)
            except OSError as e:
                # Log the error but continue with path2
                if logger.isEnabledFor(logging.DEBUG):
                    logger.debug(f'Failed to remove old icon {path1}: {e}')
            return path2
        # Old filename exists - rename to new
        try:
            rename(path1, path2)
        except OSError as e:
            # If rename fails, fall back to old path
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug(f'Failed to rename {path1} to {path2}: {e}')
            return path1
        return path2

    # Old filename does not exist - use the new one
    return path2

def macos_media_supported():
    return macos_version_at_least(10, 12, 2)

def macos_version_at_least(major, minor=0, patch=0):
    """
    Return True if the running macOS version is >= (major, minor, patch).

    If version detection fails, return False.
    """
    try:
        ver = platform.mac_ver()[0]
        if not ver:
            return False

        parts = ver.split('.')
        nums = [int(x) for x in parts]

        while len(nums) < 3:
            nums.append(0)

        return tuple(nums[:3]) >= (major, minor, patch)

    except Exception:
        return False




# -*- coding: utf-8 -*-
import glob
import os
from sys import platform

import locale
locale.setlocale(locale.LC_ALL, "")

def win_del_old_inst():
    if not platform.startswith('win'):
        return False

    path = os.getenv('PROGRAMFILES')

    out = []
    # pdirs = [f for f in glob.glob(path + "**/Python*", recursive=False)]
    pdirs = list(glob.glob(path + '**/Python*', recursive=False))

    for f in pdirs:
        # print('Searching in "' + f + '"')
        exe = os.path.join(f, 'Scripts', 'pyradio.exe')
        # # print("  EXE: {0} - {1}".format(exe, 'exists' if os.path.exists(exe) else 'not found'))
        if os.path.exists(exe):
            # print('  Found "' + exe + '"')
            out.append('DEL "' + exe + '"\n')
        site_packages = os.path.join(f, 'Lib', 'site-packages')
        if os.path.exists(site_packages):
            # # print('site-packages: "{}"'.format(site_packages))
            # pydir = [f for f in glob.glob(site_packages + "**/pyradio*.egg", recursive=False)]
            pydir = list(glob.glob(site_packages + '**/pyradio*.egg', recursive=False))
            for n in pydir:
                # print('  Found "' + n + '"')
                out.append('RD /Q /S "' + n + '"\n')
            # print(pydir)
    # print('')
    if out:
        TARGET = os.path.join(os.getenv('USERPROFILE'), 'tmp-pyradio')
        if not os.path.exists(TARGET):
            os.makedirs(TARGET)

        msg = r'''@ECHO OFF
SETLOCAL EnableDelayedExpansion

IF EXIST DEV (SET NO_DEV=0) ELSE (SET NO_DEV=1)

::net file to test privileges, 1>NUL redirects output, 2>NUL redirects errors
:: https://gist.github.com/neremin/3a4020c172053d837ab37945d81986f6
:: https://stackoverflow.com/questions/13212033/get-windows-version-in-a-batch-file
net session >nul 2>&1
IF "%errorlevel%" == "0" ( GOTO START ) ELSE (
    FOR /f "tokens=4-5 delims=. " %%i in ('ver') do SET VERSION=%%i.%%j
    IF "%version%" == "6.1" ( GOTO win7exit )
    IF "%version%" == "6.0" ( GOTO win7exit )
    GOTO getPrivileges
)

:win7exit
ECHO.
ECHO Error: You must have Administrative permissions to run this script.
ECHO        Please start cmd "As Administrator".
ECHO.
ECHO        If that does not work, ask the system administrator to
ECHO        install PyRadio for you.
ECHO.
GOTO endnopause

:getPrivileges
IF "%version%" == "6.1" ( GOTO win7exit )
IF "%version%" == "6.0" ( GOTO win7exit )

IF '%1'=='ELEV' ( GOTO START ) ELSE ( ECHO Running elevated in a different window)

SET "batchPath=%~f0"
SET "batchArgs=ELEV"

::Add quotes to the batch path, IF needed
SET "script=%0"
SET script=%script:"=%
IF '%0'=='!script!' ( GOTO PathQuotesDone )
    SET "batchPath=""%batchPath%"""
:PathQuotesDone

::Add quotes to the arguments, IF needed.
:ArgLoop
IF '%1'=='' ( GOTO EndArgLoop ) ELSE ( GOTO AddArg )
    :AddArg
    SET "arg=%1"
    SET arg=%arg:"=%
    IF '%1'=='!arg!' ( GOTO NoQuotes )
        SET "batchArgs=%batchArgs% "%1""
        GOTO QuotesDone
        :NoQuotes
        SET "batchArgs=%batchArgs% %1"
    :QuotesDone
    SHIFT
    GOTO ArgLoop
:EndArgLoop

::Create and run the vb script to elevate the batch file
ECHO SET UAC = CreateObject^("Shell.Application"^) > "%temp%\OEgetPrivileges.vbs"
ECHO UAC.ShellExecute "cmd", "/c ""!batchPath! !batchArgs!""", "", "runas", 1 >> "%temp%\OEgetPrivileges.vbs"
"%temp%\OEgetPrivileges.vbs"
EXIT /B

:START
::Remove the elevation tag and SET the correct working directory
IF '%1'=='ELEV' ( SHIFT /1 )
CD /d %~dp0

::Do your adminy thing here...

:: RD /Q /S %USERPROFILE%\tmp-pyradio
:: MKDIR %USERPROFILE%\tmp-pyradio
:: START %USERPROFILE%\tmp-pyradio


'''
        TARGET_FILE = os.path.join(TARGET, 'RunMe.bat')
        with open(TARGET_FILE, 'w', encoding='utf-8') as ofile:
            ofile.write(msg)
            ofile.write('\nECHO Looking for old style installation files\n')
            for n in out:
                ofile.write('ECHO   found ' + n.replace('RD /Q /S ', '').replace('DEL ', '').replace('\n', '') + ' - removing it...\n')
                # enable this one to not delete any files
                # ofile.write(':: ' + n)
                # enable this one to actually delete files
                ofile.write(n)
            ofile.write('ECHO.\n')
            ofile.write('ECHO Old style installation files removed...\n')
            ofile.write('ECHO.\n')
            ofile.write('PAUSE\n')

        os.startfile(TARGET_FILE)
    return True

if __name__ == '__main__':
    if not win_del_old_inst():
        print('This is a Windows only program!')

